#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "clock.h"
#include "graph.h"

//#define DEBUG
//#define FLUSH_OUT

#define EPSILON 1E-10

#define MIN_ARGS    6
#define MAX_ARGS    7
#define USAGE_STR   "Usage:  %s <input graph file> <gamma> <mu> <query vertex file> <output file> [<min clique size>]\n"
#define INPUT_ARG   1
#define GAMMA_ARG   2
#define MU_ARG      3
#define QUERY_ARG   4
#define OUTPUT_ARG  5
#define MIN_SIZE_ARG 6

#define MATRIX_CUTOFF 10000000
#define HASH_CUTOFF   10000000

#define TIME_INIT      0x1
#define TIME_PREP      0x2
#define TIME_ADD       0x4
#define TIME_REMOVE    0x8
#define TIME_HEURISTIC 0x10
#define TIME_MAX       0x20
#define TIME_OUTPUT    0x40
#define TIME_CHECK     0x80

#define TIME_CONST     (TIME_INIT | TIME_PREP)

#define HEURISTIC_DV          0x1
#define HEURISTIC_E           0x2
#define HEURISTIC_GV          0x4
#define HEURISTIC_MV          0X8
#define HEURISTIC_DIST_2      0x10
#define HEURISTIC_MAX         0x20
#define HEURISTIC_SELECT      0x40
#define HEURISTIC_CRIT_DV     0x80
#define HEURISTIC_CRIT_GV     0x100
#define HEURISTIC_CRIT_MV     0x200
#define HEURISTIC_CRIT_E      0x400
#define HEURISTIC_MIN_QC_SIZE 0x800
#define HEURISTIC_LOW_DEG     0x1000

#define HEUR_CONST        (HEURISTIC_E | HEURISTIC_MV | HEURISTIC_DV | HEURISTIC_GV | HEURISTIC_MAX | HEURISTIC_DIST_2 | HEURISTIC_CRIT_MV | HEURISTIC_CRIT_GV | HEURISTIC_CRIT_DV | HEURISTIC_CRIT_E | HEURISTIC_MIN_QC_SIZE | HEURISTIC_LOW_DEG | HEURISTIC_SELECT)

#define SELECT_GV    1
#define SELECT_MV    2
#define SELECT_DV    3
#define SELECT_SUM   4
#define SELECT_MIN   5
#define SELECT_D2    6
#define SELECT_BIG   7

#define SELECT_CONST SELECT_MIN

#define FULL_CV

#define HBI "Hierarchical bitmap index"

#define COUNT_HEUR

typedef struct
{
  graph_enum_t* v;
  #if (HEUR_CONST & HEURISTIC_GV)
  double* gv;
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  double* mv;
  #endif
  #if (HEUR_CONST & HEURISTIC_DV)
  double* dv;
  #endif
  #ifdef DEBUG
  int* nc;
  int qvc;
  #endif
  #if (HEUR_CONST & HEURISTIC_E)
  double e;
  #endif
  int s;
  int c;
  int r;
  int n;
  int size;
} Quasiclique;

//Global variables
Graph* G;
double gamma, mu;
FILE* out;
int num_qcs = 0;
int min_qc_size = 1;

Quasiclique* qc;
double prepgadjnqc, remgadjnq, addgadjnq, prepmnadjqc, remmnadjq, addmnadjq;

int add_to_qc();
void remove_from_qc();
int check_candidates(int outer);
int remove_from_c(int index, int outer);
void return_to_c(int index);
void candidate_vertex();
int is_max_mgqc();
void output_qc();
void prepare_qc(graph_enum_t vertex);
int enumerate();
void qc_swap(int index1, int index2);
#ifdef DEBUG
void print_qc();
void verify_qc();
#endif

//Timing variables
long inittime  = 0;
long preptime  = 0;
long enumtime  = 0;
long heurtime  = 0;
long addtime   = 0;
long remtime   = 0;
long maxtime   = 0;
long outtime   = 0;
long checktime = 0;

int addcount  = 0;
int remcount  = 0;
int enumcount = 0;

//Bound condition counting variables
#ifdef COUNT_HEUR
int dvcand = 0;
int ecount = 0;
int gvcand = 0;
int mvcand = 0;
int cdvcand = 0;
int cecand = 0;
int cgvcand = 0;
int cmvcand = 0;
int d2cand = 0;
int lowdegcand = 0;
int dvbound = 0;
int ebound = 0;
int gvbound = 0;
int mvbound = 0;
#endif

#if (HEUR_CONST & HEURISTIC_MAX)
#ifdef HBI
typedef unsigned char index_t;

#define BASE_LEVELS  3

int max_index(int level, int word);

int maxlevel = BASE_LEVELS;
unsigned long* delta;
#else
typedef unsigned int index_t;

#define BASE_INDEX_LENGTH 1

int index_len;
#endif
#define BRANCH_FACTOR (sizeof(index_t) << 3)

index_t* qc_index;
#endif


int main(int argc, char** argv)
{
  FILE* in;
  unsigned int query_verts;
  graph_enum_t* query;
  
  int i;
  
  //Initialize timer
  #ifdef TIME_CONST
  start_clocks();
  #if (TIME_CONST & TIME_INIT)
  select_clock(&inittime);
  #else
  select_clock(&enumtime);
  #endif
  #endif
  
  //Read in arguments
  if (argc < MIN_ARGS || argc > MAX_ARGS)
  {
    fprintf(stderr, USAGE_STR, argv[0]);
    exit(EXIT_FAILURE);
  }
  
  #ifdef DEBUG
  printf("gamma = %lf\n", gamma);
  printf("mu = %lf\n", mu);
  printf("Reading G...");
  fflush(stdout);
  #endif
  G = read_graph(argv[INPUT_ARG], MATRIX_CUTOFF, HASH_CUTOFF);
  #ifdef DEBUG
  printf("finished\n");
  printf("G has %d vertices\n", G->num_vertices);
  printf("The largest two-neighborhood of G has %d vertices\n", G->max_n2_size);
  
  /*
  for (i = 0; i < G->num_vertices; i++)
  {
    printf("N^2(%d): ", i);
    int j;
    for (j = 0; j < G->vertex[i].n2_size; j++)
      printf(" %d", G->vertex[i].n2_list[j]);
    printf("\n");
  }
  //*/
  fflush(stdout);
  #endif
  #if (HEUR_CONST & HEURISTIC_MIN_QC_SIZE)
  if (argc > MIN_SIZE_ARG)
    min_qc_size = atoi(argv[MIN_SIZE_ARG]);
  #endif

  gamma = atof(argv[GAMMA_ARG]);
  if (gamma < 0.5 | gamma > 1)
  {
    fprintf(stderr, "Gamma must be between 0.5 and 1\n");
    exit(EXIT_FAILURE);
  }
  mu = atof(argv[MU_ARG]);
  if (mu <= 0 | mu > 1)
  {
    fprintf(stderr, "Mu must be a positive real less than or equal to 1\n");
    exit(EXIT_FAILURE);
  }
  
  out = fopen(argv[OUTPUT_ARG], "w");
  if (out == NULL)
  {
    perror("Unable to open output file for writing");
    exit(EXIT_FAILURE);
  }
  
  //Allocate data structures
  qc = (Quasiclique*) malloc(sizeof(Quasiclique));
  if (qc == NULL)
  {
    perror("Error allocating main data structure");
    exit(EXIT_FAILURE);
  }
  qc->v = (graph_enum_t*) malloc(sizeof(graph_enum_t) * (G->max_n2_size + 1));
  if (qc->v == NULL)
  {
    perror("Error allocating quasiclique set");
    exit(EXIT_FAILURE);
  }
  #if (HEUR_CONST & HEURISTIC_GV)
  qc->gv = (double*) malloc(sizeof(double) * (G->max_n2_size + 1));
  if (qc->gv == NULL)
  {
    perror("Error allocating g_v array");
    exit(EXIT_FAILURE);
  }
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  qc->mv = (double*) malloc(sizeof(double) * (G->max_n2_size + 1));
  if (qc->mv == NULL)
  {
    perror("Error allocating m_v array");
    exit(EXIT_FAILURE);
  }
  #endif
  #if (HEUR_CONST & HEURISTIC_DV)
  qc->dv = (double*) malloc(sizeof(double) * (G->max_n2_size + 1));
  if (qc->dv == NULL)
  {
    perror("Error allocating d_v array");
    exit(EXIT_FAILURE);
  }
  #endif
  #ifdef DEBUG
  qc->nc = (int*) malloc(sizeof(int) * (G->max_n2_size + 1));
  if (qc->nc == NULL)
  {
    perror("Error allocating neighbor count array");
    exit(EXIT_FAILURE);
  }
  #endif
  
  #if (HEUR_CONST & HEURISTIC_MAX)
  #ifdef HBI
  delta = (unsigned long *) malloc((BASE_LEVELS + 2) * sizeof(unsigned long));
  if (delta == NULL)
  {
    perror("Error allocating helper array for maximality index");
    exit(EXIT_FAILURE);
  }
  
  #ifdef DEBUG
  printf("Delta:  1");
  fflush(stdout);
  #endif
  delta[0] = 1;
  for (i = 1; i <= BASE_LEVELS + 1; i++)
  {
    //printf("i=%d: curr val = %lu\n", i,BRANCH_FACTOR * delta[i - 1] + 1);
    delta[i] = BRANCH_FACTOR * delta[i - 1] + 1;
    #ifdef DEBUG
    printf(" %d", delta[i]);
    fflush(stdout);
    #endif
  }
  #ifdef DEBUG
  printf("\n");
  fflush(stdout);
  #endif
  
  //printf("deltabase = %lu  * num_v = %d\n",delta[BASE_LEVELS],G->num_vertices);
  //printf("together = %lu\n", delta[BASE_LEVELS]*G->num_vertices);

  qc_index = (index_t*) calloc((unsigned long)(delta[BASE_LEVELS] * G->num_vertices), sizeof(index_t));
  #else
  index_len = BASE_INDEX_LENGTH;
  
  qc_index = (index_t*) calloc(index_len * G->num_vertices, sizeof(index_t));
  #endif
  if (qc_index == NULL)
  {
    perror("Error allocating quasi-clique maximality index");
    exit(EXIT_FAILURE);
  }
  #endif
  
  //Read in query vertices
  in = fopen(argv[QUERY_ARG], "r");
  if (in == NULL)
  {
    perror("Unable to open query vertex file");
    exit(EXIT_FAILURE);
  }
  
  fscanf(in, "%u", &query_verts);

  query = (graph_enum_t*) malloc(query_verts * sizeof(graph_enum_t));
  if (query == NULL)
  {
    perror("Error allocating memory for query vertex set");
    exit(EXIT_FAILURE);
  }
  
  for (i = 0; i < query_verts; i++)
  {
    fscanf(in, "%u ", &query[i]);
    G->vertex[query[i]].active |= QUERY_VERTEX;
  }
  
  #ifdef DEBUG
  printf("Q has %d vertices: ", query_verts);
  for (i = 0; i < query_verts; i++)
    printf(" %d", query[i]);
  printf("\n");
  fflush(stdout);
  #endif
  
  fclose(in);
  
  #if (HEUR_CONST & (HEURISTIC_GV | HEURISTIC_MV))
  //Define g_v and m_v constants
  if (1 - mu - gamma > 0)
  {
    #if (HEUR_CONST & HEURISTIC_GV)
    prepgadjnqc = 1 - mu - gamma;
    remgadjnq = 1 - mu - gamma;
    addgadjnq = 0;
    #endif
    #if (HEUR_CONST & HEURISTIC_MV)
    prepmnadjqc = 1 - mu - gamma;
    remmnadjq = 1 - mu - gamma;
    addmnadjq = 0;
    #endif
  }
  else
  {
    #if (HEUR_CONST & HEURISTIC_GV)
    prepgadjnqc = 0;
    remgadjnq = 0;
    addgadjnq = -(1 - mu - gamma);
    #endif
    #if (HEUR_CONST & HEURISTIC_MV)
    prepmnadjqc = 0;
    remmnadjq = 0;
    addmnadjq = -(1 - mu - gamma);
    #endif
  }
  #endif
  
  #ifdef TIME_CONST
  select_clock(&enumtime);
  #endif
  
  for (i = 0; i < query_verts; i++)
  {
    #ifdef DEBUG
    printf("Vertex #%d (%d)\n", i, query[i]);
    fflush(stdout);
    #endif
    prepare_qc(query[i]);
    #ifdef DEBUG
    verify_qc();
    #endif
    #if (HEUR_CONST & HEURISTIC_MAX)
    //if (qc->r < qc->n && remove_from_c(qc->r, -1) == 0 || check_maximality() == 0)
    if (check_maximality() == 0)
    {
      #ifdef DEBUG
      printf("Vertex failed\n");
      fflush(stdout);
      #endif
      continue;
    }
    #endif
    enumerate();
  }
  
  fclose(out);
  
  //Output results
  #ifdef DEBUG
  printf("\n");
  fflush(stdout);
  #endif
  printf("%d mu, gamma-quasi-cliques discovered!\n\n", num_qcs);
  
  printf("Function summary\n-----------------\nEnumerate(): %8d calls\nAdd():       %8d calls\nRemove():    %8d calls\n\n", enumcount, addcount, remcount);
  
  //Output timing information (if applicable)
  #ifdef TIME_CONST
  stop_clocks();
  printf("Timing summary (1000 cycles)\n----------------------------\n");
  #if (TIME_CONST & TIME_INIT)
  printf("Initialization         %8ld\n", inittime / 1000);
  #endif
  #if (TIME_CONST & TIME_PREP)
  printf("Preparing S            %8ld\n", preptime / 1000);
  #endif
  printf("Algorithm              %8ld\n", enumtime / 1000);
  #if (TIME_CONST & TIME_HEURISTIC)
  printf("Heuristic              %8ld\n", heurtime / 1000);
  #endif
  #if (TIME_CONST & TIME_ADD)
  printf("Adding vertices        %8ld\n", addtime / 1000);
  #endif
  #if (TIME_CONST & TIME_REMOVE)
  printf("Removing vertices      %8ld\n", remtime / 1000);
  #endif
  #if (TIME_CONST & TIME_CHECK)
  printf("Candidate set checking %8ld\n", checktime / 1000);
  #endif
  #if (TIME_CONST & TIME_MAX)
  printf("Maximality checking    %8ld\n", maxtime / 1000);
  #endif
  #if (TIME_CONST & TIME_OUTPUT)
  printf("Output                 %8ld\n", outtime / 1000);
  #endif
  printf("\nTotal                  %8ld\n", (inittime + preptime + enumtime
       + heurtime + addtime + remtime + maxtime + outtime) / 1000);
  #endif
  
  //Output bound condition information
  #ifdef COUNT_HEUR
  printf("\nHeuristic summary (Candidates removed / # of times backtracking)\n----------------------------\n");
  #if (HEUR_CONST & HEURISTIC_DV)
  printf("dv       %7d\t%7d\n", dvcand, dvbound);
  #if (HEUR_CONST & HEURISTIC_CRIT_DV)
  printf("crit dv  %7d\n", cdvcand);
  #endif
  #endif
  #if (HEUR_CONST & HEURISTIC_E)
  printf("e        %7d\t%7d\n", ecount, ebound);
  #if (HEUR_CONST & HEURISTIC_CRIT_E)
  printf("crit e   %7d\n", cecand);
  #endif
  #endif
  #if (HEUR_CONST & HEURISTIC_GV)
  printf("gv       %7d\t%7d\n", gvcand, gvbound);
  #if (HEUR_CONST & HEURISTIC_CRIT_GV)
  printf("crit gv  %7d\n", cgvcand);
  #endif
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  printf("mv       %7d\t%7d\n", mvcand, mvbound);
  #if (HEUR_CONST & HEURISTIC_CRIT_MV)
  printf("crit mv  %7d\n", cmvcand);
  #endif
  #endif
  #if (HEUR_CONST & HEURISTIC_DIST_2)
  printf("degree 2 %7d\n", d2cand);
  #endif
  #if (HEUR_CONST & HEURISTIC_LOW_DEG)
  printf("low deg  %7d\n", lowdegcand);
  #endif
  #endif
  
  free_graph(G);
  free(qc->v);
  #if (HEUR_CONST & HEURISTIC_DV)
  free(qc->dv);
  #endif
  #if (HEUR_CONST & HEURISTIC_GCV)
  free(qc->gv);
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  free(qc->mv);
  #endif
  #ifdef DEBUG
  free(qc->nc);
  #endif
  free(qc);
  free(query);
  
  #if (HEUR_CONST & HEURISTIC_MAX)
  #ifdef HBI
  free(delta);
  #endif
  free(qc_index);
  #endif
  
  exit(EXIT_SUCCESS);
}


#ifdef DEBUG
void print_qc()
{
  int i = 0;
  
  printf("%d", qc->v[i]);
  for (i++; i < qc->c; i++)
    printf((G->vertex[qc->v[i]].active & TO_REMOVE)? " (%d)":" %d", qc->v[i]);
  printf(" |");
  //for (; i < qc->r; i++)
  //  printf((G->vertex[qc->v[i]].active & TO_REMOVE)? " (%d)":" %d", qc->v[i]);
  //printf(" (");
  for (; i < qc->n; i++)
    printf((G->vertex[qc->v[i]].active & TO_REMOVE)? " (%d)":" %d", qc->v[i]);
  printf(" |");
  //printf(" ) |");
  for (; i < qc->size; i++)
    printf((G->vertex[qc->v[i]].active & TO_REMOVE)? " (%d)":" %d", qc->v[i]);
  printf("\n");
  
  //return;
  printf(" %d\n", qc->qvc);
  
  printf(" %d", qc->nc[0]);
  for (i = 1; i < qc->size; i++)
  {
    if (i == qc->c)
      printf(" |");
    //if (i == qc->r)
    //  printf(" (");
    if (i == qc->n)
      printf(" |");
    //  printf(" ) |");
    printf(" %d", qc->nc[i]);
  }
  //if (qc->r == qc->size)
  //  printf(" )\n");
  //else
    printf("\n");
  #if (HEUR_CONST & HEURISTIC_E)
  printf(" %lf\n", qc->e);
  #endif
  
  #if (HEUR_CONST & HEURISTIC_GV)
  for (i = 0; i < qc->c; i++)
    printf(" %lf", qc->gv[i]);
  #ifdef FULL_CV
  printf(" |");
  for (; i < qc->n; i++)
    printf(" %lf", qc->gv[i]);
  #endif
  printf("\n");
  #endif
  
  #if (HEUR_CONST & HEURISTIC_MV)
  for (i = 0; i < qc->c; i++)
    printf(" %lf", qc->mv[i]);
  #ifdef FULL_CV
  printf(" |");
  for (; i < qc->n; i++)
    printf(" %lf", qc->mv[i]);
  #endif
  printf("\n");
  #endif
  
  #if (HEUR_CONST & HEURISTIC_DV)
  printf(" %lf", qc->dv[0]);
  for (i = 1; i < qc->size; i++)
  {
    if (i == qc->c)
      printf(" |");
    if (i == qc->n)
      printf(" |");
    printf(" %lf", qc->dv[i]);
  }
  printf("\n");
  #endif
  fflush(stdout);
}

int is_qc()
{
  int i;
  
  if (qc->n > qc->c)
  {
    //Check mu condition
    if (qc->qvc < mu * qc->c + EPSILON)
    {
      printf("Failed mu condition\n");
      return 0;
    }
    
    //Check gamma condition
    for (i = 0; i < qc->c; i++)
      if (qc->nc[i] < gamma * (qc->c - 1) - EPSILON)
        {
          printf("Failed gamma condition on vertex %d (%d vs. %lf)\n", qc->v[i], qc->nc[i], gamma * (qc->c - 1) - EPSILON);
          return 0;
        }
  }
  
  printf("Mu, gamma-quasi-clique!\n");
  return -1;
}
#endif

#ifdef OLD_MAXIMALITY
int is_max_mgqc()
{
  #if (TIME_CONST & TIME_MAXIMALITY)
  select_clock(&maxtime);
  #endif
  
  #ifdef HBI
  fprintf(stderr, "Should not call!\n");
  exit(EXIT_FAILURE);
  #endif
  /*
  int i, j, old_e;
  old_e = qc->e;
  for (i = qc->n; i < qc->size; i++)
  {
    if (G->vertex[qc->v[i]].active & TO_REMOVE)
      continue;
    if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
      qc->e += 1 - mu;
    
    qc->cv[i] = qc->dv[i];
    qc->dv[i] = -gamma;
    for (j = 0; j < qc->c; j++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        qc->dv[i] += 1 - gamma;
        qc->dv[j] += 1 - gamma;
      }
    for (j = i + 1; j < qc->size; j++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        qc->dv[i] += 1 - gamma;
        qc->dv[j] += 1 - gamma;
      }
  }
  //*/
  
  int i, temp, old_n, old_nc;
  int max_qc = -1;
  
  #ifdef DEBUG
  printf("Maximality check\n");
  fflush(stdout);
  #endif
  
  //Check maximality
  old_n = qc->n;
  for (; max_qc && qc->n < qc->size; qc->n++)
  {
    //Max heuristic??
    if (G->vertex[qc->v[qc->n]].active & TO_REMOVE)
    {
      #ifdef DEBUG
      printf("Skipping vertex %d\n", qc->v[qc->n]);
      fflush(stdout);
      #endif
      continue;
    }
    
    //"Add" qc->v[qc->n] to S
    #ifdef DEBUG
    printf("Adding vertex %d: ", qc->v[qc->n]);
    fflush(stdout);
    #endif
    old_nc = qc->nc[qc->n];
    qc->nc[qc->n] = qc->nc[qc->c];
    qc->nc[qc->c] = 0;
    for (i = 0; i < qc->c; i++)
      if (connected(G, qc->v[qc->n], qc->v[i]))
      {
        qc->nc[i]++;
        qc->nc[qc->c]++;
      }
      else if (two_connected(G, qc->v[qc->n], qc->v[i]) == 0)
      {
        for (i--; i >= 0; i--)
          if (connected(G, qc->v[qc->n], qc->v[i]))
            qc->nc[i]--;
        qc->nc[qc->c] = qc->nc[qc->n];
        qc->nc[qc->n] = old_nc;
        #ifdef DEBUG
        printf("failed!\n");
        fflush(stdout);
        #endif
        break;
      }
    if (i < qc->c)
      continue;
    if (G->vertex[qc->v[qc->n]].active & QUERY_VERTEX)
      qc->qvc++;
    temp = qc->v[qc->n];
    qc->v[qc->n] = qc->v[qc->c];
    qc->v[qc->c] = temp;
    #ifdef DEBUG
    qc->n++;
    qc->c++;
    print_qc();
    qc->c--;
    qc->n--;
    #endif
    
    //Check if qc
    if (qc->qvc >= mu * qc->c + EPSILON)
    {
      #ifdef DEBUG
      printf("Passed mu check\n");
      fflush(stdout);
      #endif
      for (i = 0; i <= qc->c; i++)
        if (qc->nc[i] < gamma * (qc->c - 1) - EPSILON)
        {
          #ifdef DEBUG
          printf("Failed gamma check on vertex %d\n", qc->v[i]);
          fflush(stdout);
          #endif
          break;
        }
      if (i > qc->c)
      {
        #ifdef DEBUG
        printf("Passed gamma check\n");
        fflush(stdout);
        #endif
        max_qc = 0;
      }
    }
    #ifdef DEBUG
    else
      printf("Failed mu check\n");
    printf("%s, gamma-quasi-clique\n", max_qc? "Not a mu":"Mu");
    fflush(stdout);
    #endif
    if (max_qc && qc->n < qc->size)
    {
      qc->c++;
      max_qc = is_max_mgqc();
      qc->c--;
    }
    
    #if (TIME_CONST & TIME_MAXIMALITY)
    select_clock(&maxtime);
    #endif
    
    //"Remove" qc->v[qc->n] from S
    qc->v[qc->c] = qc->v[qc->n];
    qc->v[qc->n] = temp;
    qc->nc[qc->c] = qc->nc[qc->n];
    qc->nc[qc->n] = old_nc;
    for (i = 0; i < qc->c; i++)
      if (connected(G, temp, qc->v[i]))
        qc->nc[i]--;
    if (G->vertex[temp].active & QUERY_VERTEX)
      qc->qvc++;
    
    #ifdef DEBUG
    printf("Removing vertex %d: ", temp);
    print_qc();
    fflush(stdout);
    #endif
  }
  qc->n = old_n;
  
  #if (TIME_CONST & TIME_MAXIMALITY)
  select_clock(&enumtime);
  #endif
  
  return max_qc;
}
#endif

#ifdef DEBUG
void verify_qc()
{
  int i, j, nc, vertex;
  double dv, gv, mv;
  for (i = 0; i < qc->n; i++)
  {
    nc = 0;
    dv = 0;
    vertex = qc->v[i];
    #if (HEUR_CONST & HEURISTIC_GV)
    gv = -gamma * ((G->vertex[vertex].active & QUERY_VERTEX) == 0) + gamma * (1 - mu);
    #endif
    #if (HEUR_CONST & HEURISTIC_MV)
    mv = (((G->vertex[vertex].active & QUERY_VERTEX) != 0) - mu) * (1 - gamma);
    #endif
    for (j = 0; j < qc->c; j++)
    {
      if (i == j)
        continue;
      if (connected(G, vertex, qc->v[j]))
      {
        nc++;
        #if (HEUR_CONST & HEURISTIC_DV)
        dv += 1 - gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_GV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          gv += 1 - mu;
        else
          gv += 1 - mu - gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          mv += 1 - gamma;
        #endif
      }
      else
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        dv -= gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_GV)
        if ((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0)
          gv -= gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          mv += 1 - mu - gamma;
        else
          mv -= mu;
        #endif
      }
    }
    for (; j < qc->n; j++)
    {
      if (i == j)
        continue;
      if (connected(G, vertex, qc->v[j]))
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        dv += 1 - gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_GV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          gv += 1 - mu;
        else
          gv += prepgadjnqc;
        #endif
        #if (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          mv += 1 - gamma;
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV)
      else if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
        mv += prepmnadjqc;
      #endif
    }
    
    if (nc != qc->nc[i])
    {
      printf("NC[%d] = %d (!= %d)\n", qc->v[i], qc->nc[i], nc);
      print_qc();
      exit(EXIT_FAILURE);
    }
    #if (HEUR_CONST & HEURISTIC_DV)
    if ((dv + EPSILON < qc->dv[i]) | (dv - EPSILON > qc->dv[i]))
    {
      printf("D_V[%d] = %lf (!= %lf)\n", qc->v[i], qc->dv[i], dv);
      print_qc();
      exit(EXIT_FAILURE);
    }
    #endif
    
    #if (HEUR_CONST & HEURISTIC_GV)
    #ifndef FULL_CV
    if (i < qc->c)
    #endif
      if ((gv + EPSILON < qc->gv[i]) | (gv - EPSILON > qc->gv[i]))
      {
        printf("G_V[%d] = %lf (!= %lf)\n", qc->v[i], qc->gv[i], gv);
        print_qc();
        exit(EXIT_FAILURE);
      }
    #endif
    
    #if (HEUR_CONST & HEURISTIC_MV)
    #ifndef FULL_CV
    if (i < qc->c)
    #endif
      if ((mv + EPSILON < qc->mv[i]) | (mv - EPSILON > qc->mv[i]))
      {
        printf("m_V[%d] = %lf (!= %lf)\n", qc->v[i], qc->mv[i], mv);
        print_qc();
        exit(EXIT_FAILURE);
      }
    #endif
  }
}
#endif

void candidate_vertex()
{
  // Begin heuristic
  #if (TIME_CONST & TIME_HEURISTIC)
  select_clock(&heurtime);
  #endif
  
  int i;
  
  #if ((HEUR_CONST & HEURISTIC_SELECT) == HEURISTIC_SELECT)
  #if (SELECT_CONST == SELECT_BIG)
  fprintf(stderr, "Error:  BIG heuristic no longer supported.\n");
  exit(1);
  /*
  int index = 0;
  double critical = DBL_MAX;
  
  int sel = -1;
  
  for (i = 0; i < qc->c; i++)
  {
    #if (HEUR_CONST & HEURISTIC_DV)
    if (qc->dv[i] < critical)
    {
      critical = qc->dv[i];
      index = i << 1;
    }
    #endif
    #if (HEUR_CONST & HEURISTIC_GV)
    if (qc->gv[i] < critical)
    {
      critical = qc->gv[i];
      index = (i << 1) + 1;
    }
    #endif
    #if (HEUR_CONST & HEURISTIC_MV)
    if (qc->mv[i] < critical)
    {
      critical = qc->mv[i];
      index = (i << 1) + 1;
    }
    #endif
  }
  
  #if (HEUR_CONST & HEURISTIC_E)
  if (qc->e * gamma / mu < critical)
  {
    critical = qc->e * gamma / mu;
    for (; i < qc->n; i++)
    {
      if ((G->vertex[qc->v[i]].active & QUERY_VERTEX) == 0)
        sel = i;
      #if (HEUR_CONST & HEURISTIC_DV)
      if (qc->dv[i] < critical)
      {
        critical = qc->dv[i];
        sel = i;
        for (i++; i < qc->n; i++)
        {
          if (qc->dv[i] < critical)
          {
            critical = qc->dv[i];
            sel = i;
          }
          #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
          if (qc->gv[i] < critical)
          {
            critical = qc->gv[i];
            sel = i;
          }
          #endif
          #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
          if (qc->mv[i] < critical)
          {
            critical = qc->mv[i];
            sel = i;
          }
          #endif
        }
        break;
      }
      #endif
      #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
      if (qc->gv[i] < critical)
      {
        critical = qc->gv[i];
        sel = i;
        for (i++; i < qc->n; i++)
        {
          #if (HEUR_CONST & HEURISTIC_DV)
          if (qc->dv[i] < critical)
          {
            critical = qc->dv[i];
            sel = i;
          }
          #endif
          if (qc->gv[i] < critical)
          {
            critical = qc->gv[i];
            sel = i;
          }
        }
      }
      #endif
    }
  }
  else
  #endif
  #if (HEUR_CONST & HEURISTIC_GV)
  if (index & 1)
  {
    index >>= 1;
    for (; i < qc->n; i++)
    {
      if ((G->vertex[qc->v[i]].active & QUERY_VERTEX) == 0
           && connected(G, qc->v[index], qc->v[i]) == 0)
        sel = i;
      #if (HEUR_CONST & HEURISTIC_DV)
      if (qc->dv[i] < critical)
      {
        critical = qc->dv[i];
        sel = i;
        for (i++; i < qc->n; i++)
        {
          if (qc->dv[i] < critical)
          {
            critical = qc->dv[i];
            sel = i;
          }
          #ifdef FULL_CV
          if (qc->gv[i] < critical)
          {
            critical = qc->gv[i];
            sel = i;
          }
          #endif
        }
        break;
      }
      #endif
      #ifdef FULL_CV
      if (qc->gv[i] < critical)
      {
        critical = qc->gv[i];
        sel = i;
        for (i++; i < qc->n; i++)
        {
          #if (HEUR_CONST & HEURISTIC_DV)
          if (qc->dv[i] < critical)
          {
            critical = qc->dv[i];
            sel = i;
          }
          #endif
          if (qc->gv[i] < critical)
          {
            critical = qc->gv[i];
            sel = i;
          }
        }
        break;
      }
      #endif
    }
  }
  else
  #endif
  #if (HEUR_CONST & HEURISTIC_DV)
  {
    index >>= 1;
    for (; i < qc->n; i++)
    {
      if (connected(G, qc->v[index], qc->v[i]) == 0)
        sel = i;
      if (qc->dv[i] < critical)
      {
        critical = qc->dv[i];
        sel = i;
        for (i++; i < qc->n; i++)
        {
          if (qc->dv[i] < critical)
          {
            critical = qc->dv[i];
            sel = i;
          }
          #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
          if (qc->gv[i] < critical)
          {
            critical = qc->gv[i];
            sel = i;
          }
          #endif
        }
        break;
      }
    }
  }
  #else
    ;
  #endif
  
  if (sel == -1)
  {
    sel = qc->c;
    #if (HEUR_CONST & HEURISTIC_CV && defined(FULL_CV))
    critical = qc->cv[qc->c];
    for (i = qc->c + 1; i < qc->n; i++)
      if (qc->cv[i] < critical)
      {
        critical = qc->cv[i];
        sel = i;
      }
    #elif (HEUR_CONST & HEURISTIC_DV)
    critical = qc->dv[qc->c];
    for (i = qc->c + 1; i < qc->n; i++)
      if (qc->dv[i] < critical)
      {
        critical = qc->dv[i];
        sel = i;
      }
    #elif (defined(DEBUG))
    critical = qc->nc[qc->c];
    for (i = qc->c + 1; i < qc->n; i++)
      if (qc->nc[i] < critical)
      {
        critical = qc->nc[i];
        sel = i;
      }
    #endif
  }
  //*/
  #else
  int sel = qc->c;
  double critical;
  #if ((SELECT_CONST == SELECT_DV) | (SELECT_CONST == SELECT_MIN) | (SELEcT_CONST == SELECT_SUM))
  critical = qc->dv[qc->c];
  #if (SELECT_CONST == SELECT_MIN && defined(FULL_CV))
  #if (HEUR_CONST & HEURISTIC_GV)
  if (qc->gv[qc->c] < critical)
    critical = qc->gv[qc->c];
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  if (qc->mv[qc->c] < critical)
    critical = qc->mv[qc->c];
  #endif
  #endif
  #elif (SELECT_CONST == SELECT_SUM)
  #if (HEUR_CONST & HEURISTIC_GV)
  critical += qc->gv[qc->c];
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  critical += qc->mv[qc->c];
  #endif
  #elif (((HEUR_CONST & HEURISTIC_GV) == HEURISTIC_GV) && (SELECT_CONST == SELECT_GV))
  critical = qc->gv[qc->c];
  #elif (((HEUR_CONST & HEURISTIC_MV) == HEURISTIC_MV) && (SELECT_CONST == SELECT_MV))
  critical = qc->mv[qc->c];
  #elif (SELECT_CONST == SELECT_D2)
  critical = G->vertex[qc->v[qc->c]].n2_size;
  #endif
  for (i = qc->c + 1; i < qc->n; i++)
  {
    double val;
    #if ((SELECT_CONST == SELECT_DV) | (SELECT_CONST == SELECT_MIN) | (SELECT_CONST == SELECT_SUM))
    val = qc->dv[i];
    #elif (SELECT_CONST == SELECT_GV)
    #if (HEUR_CONST & HEURISTIC_GV)
    val = qc->gv[i];
    #endif
    #elif (SELECT_CONST == SELECT_D2)
    val = G->vertex[qc->v[i]].n2_size;
    #endif
    #if (SELECT_CONST == SELECT_SUM)
    #if (HEUR_CONST & HEURISTIC_GV)
    val += qc->gv[i];
    #endif
    #if (HEUR_CONST & HEURISTIC_MV)
    val += qc->mv[i];
    #endif
    #endif
    if (val < critical)
    {
      critical = val;
      sel = i;
    }
    #if (SELECT_CONST == SELECT_MIN)
    #if (HEUR_CONST & HEURISTIC_GV)
    if (qc->gv[i] < critical)
    {
      critical = qc->gv[i];
      sel = i;
    }
    #endif
    #if (HEUR_CONST & HEURISTIC_MV)
    if (qc->mv[i] < critical)
    {
      critical = qc->mv[i];
      sel = i;
    }
    #endif
    #endif
  }
  #endif
  
  qc_swap(qc->c, sel);
  #endif
}

int add_to_qc()
{
  #if (TIME_CONST & TIME_ADD)
  select_clock(&addtime);
  #elif (TIME_CONST & TIME_HEURISTIC)
  select_clock(&enumtime);
  #endif
  addcount++;
  
  int i, j;
  int success = -1;
  graph_enum_t vertex = qc->v[qc->c];
  int active = G->vertex[vertex].active;
  
  #ifdef HALF_CV
  #if (HEUR_CONST & HEURISTIC_GV)
  qc->gv[qc->c] = gamma * (1 - mu);
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  qc->mv[qc->c] = mu * (gamma - 1);
  #endif
  #endif
  if ((active & QUERY_VERTEX) == 0)
  {
    #if (HEUR_CONST & HEURISTIC_GV && defined(HALF_CV))
    qc->gv[qc->c] -= gamma;
    #endif
    #if (HEUR_CONST & HEURISTIC_E)
    qc->e -= mu;
    success &= (qc->e >= -EPSILON);
    #ifdef COUNT_HEUR
    if (qc->e < -EPSILON)
      ebound++;
    #endif
    #if (HEUR_CONST & HEURISTIC_CRIT_E)
    if (qc->e < mu - EPSILON)
    {
      #ifdef DEBUG
      printf("Mu is critical!\n");
      fflush(stdout);
      #endif
      //for (j = qc->n - 1; j > qc->c; j--)
      for (j = qc->c + 1; j < qc->n; j++)
        if ((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0)
        {
          //qc_swap(j, --qc->r);
          #ifdef COUNT_HEUR
          cecand++;
          #endif
          G->vertex[qc->v[j]].active |= TO_REMOVE;
        }
    }
    #endif
    #endif
  }
  else
  {
    #ifdef DEBUG
    qc->qvc++;
    #endif
    #if (HEUR_CONST & HEURISTIC_MV && defined(HALF_CV))
    qc->mv[qc->c] += 1 - gamma;
    #endif
  }
  
  for (i = 0; i < qc->c; i++)
  {
    if (connected(G, qc->v[i], vertex))
    {
      #ifdef DEBUG
      qc->nc[i]++;
      #endif
      #ifdef HALF_CV
      #if (HEUR_CONST & HEURISTIC_GV)
      qc->gv[qc->c] += 1 - mu;
      #endif
      #if (HEUR_CONST & HEURISTIC_MV)
      if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        qc->mv[qc->c] += 1 - gamma;
      else
        qc->mv[qc->c] -= mu;
      #endif
      #endif
      #if (HEUR_CONST & HEURISTIC_GV)
      if ((active & QUERY_VERTEX) == 0) //R
      {
        qc->gv[i] -= addgadjnq;
        success &= (qc->gv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->gv[i] < -EPSILON)
          gvbound++;
        #endif
        #if (HEUR_CONST & HEURISTIC_CRIT_GV)
        if (success & (qc->gv[i] < gamma - EPSILON))
        {
          #ifdef DEBUG
          printf("Vertex %d has a critical gv (-N - Q)\n", qc->v[i]);
          fflush(stdout);
          #endif
          //for (j = qc->n - 1; j > qc->c; j--)
          for (j = qc->c + 1; j < qc->n; j++)
            if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]) == 0)
            {
              //qc_swap(j, --qc->r);
              #ifdef COUNT_HEUR
              cgvcand++;
              #endif
              G->vertex[qc->v[j]].active |= TO_REMOVE;
            }
        }
        if (success & (qc->gv[i] < addgadjnq - EPSILON))
        {
          #ifdef DEBUG
          printf("Vertex %d has a critical gv (N - Q)\n", qc->v[i]);
          fflush(stdout);
          #endif
          //for (j = qc->n - 1; j > qc->c; j--)
          for (j = qc->c + 1; j < qc->n; j++)
            if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]))
            {
              #ifdef COUNT_HEUR
              cgvcand++;
              #endif
              //qc_swap(j, --qc->r);
              G->vertex[qc->v[j]].active |= TO_REMOVE;
            }
        }
        #endif
      }
      #endif
    }
    else
    {
      if ((active & QUERY_VERTEX) == 0)
      {
        #if (HEUR_CONST & HEURISTIC_GV)
        qc->gv[i] -= gamma;
        success &= (qc->gv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->gv[i] < -EPSILON)
          gvbound++;
        #endif
        #endif
        #if (HEUR_CONST & HEURISTIC_MV)
        qc->mv[i] -= mu;
        success &= (qc->mv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->mv[i] < -EPSILON)
          mvbound++;
        #endif
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV)
      else
      {
        qc->mv[i] -= addmnadjq;
        success &= (qc->mv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->mv[i] < -EPSILON)
          mvbound++;
        #endif

      }
      #endif
      #ifdef HALF_CV
      #if (HEUR_CONST & HEURISTIC_GV)
      qc->gv[qc->c] -= gamma;
      #endif
      #if (HEUR_CONST & HEURISTIC_MV)
      if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        qc->mv[qc->c] += 1 - gamma;
      else
        qc->mv[qc->c] -= mu;
      #endif
      #endif
      #if (HEUR_CONST & HEURISTIC_CRIT_GV)
      if (success & (qc->gv[i] < gamma - EPSILON))
      {
        #ifdef DEBUG
        printf("Vertex %d has a critical gv (-N - Q)\n", qc->v[i]);
        fflush(stdout);
        #endif
        //for (j = qc->n - 1; j > qc->c; j--)
        for (j = qc->c + 1; j < qc->n; j++)
          if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]) == 0)
          {
            #ifdef COUNT_HEUR
            cgvcand++;
            #endif
            //qc_swap(j, --qc->r);
            G->vertex[qc->v[j]].active |= TO_REMOVE;
          }
      }
      if (success & (qc->gv[i] < addgadjnq - EPSILON))
      {
        #ifdef DEBUG
        printf("Vertex %d has a critical gv (N - Q)\n", qc->v[i]);
        fflush(stdout);
        #endif
        //for (j = qc->n - 1; j > qc->c; j--)
        for (j = qc->c + 1; j < qc->n; j++)
          if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]))
          {
            #ifdef COUNT_HEUR
            cgvcand++;
            #endif
            //qc_swap(j, --qc->r);
            G->vertex[qc->v[j]].active |= TO_REMOVE;
          }
      }
      #endif
      #if (HEUR_CONST & HEURISTIC_CRIT_MV)
      if (success & (qc->mv[i] < mu - EPSILON))
      {
        #ifdef DEBUG
        printf("Vertex %d has a critical mv (-N - Q)\n", qc->v[i]);
        fflush(stdout);
        #endif
        //for (j = qc->n - 1; j > qc->c; j--)
        for (j = qc->c + 1; j < qc->n; j++)
          if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]) == 0)
          {
            #ifdef COUNT_HEUR
            cmvcand++;
            #endif
            //qc_swap(j, --qc->r);
            G->vertex[qc->v[j]].active |= TO_REMOVE;
          }
      }
      if (success & (qc->mv[i] < addmnadjq - EPSILON))
      {
        #ifdef DEBUG
        printf("Vertex %d has a critical mv (Q - N)\n", qc->v[i]);
        fflush(stdout);
        #endif
        //for (j = qc->n - 1; j > qc->c; j--)
        for (j = qc->c + 1; j < qc->n; j++)
          if (G->vertex[qc->v[j]].active & QUERY_VERTEX && (connected(G, qc->v[i], qc->v[j]) == 0))
          {
            #ifdef COUNT_HEUR
            cmvcand++;
            #endif
            //qc_swap(j, --qc->r);
            G->vertex[qc->v[j]].active |= TO_REMOVE;
          }
      }
      #endif
      //#endif
      
      #if (HEUR_CONST & HEURISTIC_DV)
      qc->dv[i] -= gamma;
      success &= (qc->dv[i] >= -EPSILON);
      #ifdef COUNT_HEUR
      if (qc->dv[i] < -EPSILON)
        dvbound++;
      #endif
      #if (HEUR_CONST & HEURISTIC_CRIT_DV)
      if (success & (qc->dv[i] < gamma - EPSILON))
      {
        #ifdef DEBUG
        printf("Vertex %d has a critical dv\n", qc->v[i]);
        fflush(stdout);
        #endif
        //for (j = qc->n - 1; j > qc->c; j--)
        for (j = qc->c + 1; j < qc->n; j++)
          if (connected(G, qc->v[i], qc->v[j]) == 0)
          {
            #ifdef COUNT_HEUR
            cdvcand++;
            #endif
            //qc_swap(j, --qc->r);
            G->vertex[qc->v[j]].active |= TO_REMOVE;
          }
      }
      #endif
      #endif
    }
  }
  
  if (success)
  {
    //for (i = qc->n - 1; i > qc->c; i--)
    for (i = qc->c + 1; i < qc->n; i++)
    {
      #if (HEUR_CONST & HEURISTIC_DIST_2)
      if (two_connected(G, qc->v[i], vertex) == 0)
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[i] -= gamma;
        #endif
        #ifdef FULL_CV
        #if (HEUR_CONST & HEURISTIC_MV)
        if ((active & QUERY_VERTEX) == 0)
        {
          qc->mv[i] -= mu;
          #if (HEUR_CONST & HEURISTIC_GV)
          qc->gv[i] -= gamma;
          #endif
        }
        else
          qc->mv[i] -= addmnadjq;
        #elif (HEUR_CONST & HEURISTIC_GV)
        if ((active & QUERY_VERTEX) == 0)
          qc->gv[i] -= gamma;
        #endif
        #endif
        #ifdef DEBUG
        printf("Vertex %d is not 2-connected to %d\n", qc->v[i], vertex);
        fflush(stdout);
        #endif
        #ifdef COUNT_HEUR
        d2cand++;
        #endif
        G->vertex[qc->v[i]].active |= TO_REMOVE;
        //qc_swap(i, --qc->r);
        
        #if (HEUR_CONST & HEURISTIC_MV && defined(HALF_CV))
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
          qc->mv[qc->c] += prepmnadjqc;
        #endif
      }
      else
      #endif
      if (connected(G, qc->v[i], vertex) == 0)
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[i] -= gamma;
        if (qc->dv[i] < -EPSILON)
        {
          #ifdef DEBUG
          printf("Vertex %d has dv < 0\n", qc->v[i]);
          fflush(stdout);
          #endif
          #ifdef COUNT_HEUR
          dvcand++;
          #endif
          G->vertex[qc->v[i]].active |= TO_REMOVE;
          //qc_swap(i, --qc->r);
        }
        #endif
        #ifdef FULL_CV
        #if (HEUR_CONST & HEURISTIC_MV)
        if ((active & QUERY_VERTEX) == 0)
        {
          qc->mv[i] -= mu;
          #if (HEUR_CONST & HEURISTIC_GV)
          qc->gv[i] -= gamma;
          #endif
        }
        else
          qc->mv[i] -= addmnadjq;
        #elif (HEUR_CONST & HEURISTIC_GV)
        if ((active & QUERY_VERTEX) == 0)
          qc->gv[i] -= gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_GV)
        if (qc->gv[i] < -EPSILON)
        {
          #ifdef DEBUG
          printf("Vertex %d has gv < 0\n", qc->v[i]);
          fflush(stdout);
          #endif
          #ifdef COUNT_HEUR
          gvcand++;
          #endif
          G->vertex[qc->v[i]].active |= TO_REMOVE;
          //qc_swap(i, --qc->r);
        }
        #endif
        #if (HEUR_CONST & HEURISTIC_MV)
        if (qc->mv[i] < -EPSILON)
        {
          #ifdef DEBUG
          printf("Vertex %d has mv < 0\n", qc->v[i]);
          fflush(stdout);
          #endif
          #ifdef COUNT_HEUR
          mvcand++;
          #endif
          G->vertex[qc->v[i]].active |= TO_REMOVE;
          //qc_swap(i, --qc->r);
        }
        #endif
        #endif
        #ifdef HALF_CV
        #if (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
          qc->mv[qc->c] += prepmnadjqc;
        #endif
        #endif
      }
      else
      {
        #ifdef DEBUG
        qc->nc[i]++;
        #endif
        #ifdef FULL_CV
        #if (HEUR_CONST & HEURISTIC_GV)
        if ((active & QUERY_VERTEX) == 0)
          qc->gv[i] -= addgadjnq;
        if (qc->gv[i] < -EPSILON)
        {
          #ifdef DEBUG
          printf("Vertex %d has gv < 0\n", qc->v[i]);
          fflush(stdout);
          #endif
          #ifdef COUNT_HEUR
          gvcand++;
          #endif
          G->vertex[qc->v[i]].active |= TO_REMOVE;
          //qc_swap(i, --qc->r);
        }
        #endif
        #endif
        #ifdef HALF_CV
        #if (HEUR_CONST & HEURISTIC_GV)
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        {
          qc->gv[qc->c] += 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV)
          qc->mv[qc->c] += 1 - gamma;
          #endif
        }
        else
          qc->gv[qc->c] += prepgadjnqc;
        #elif (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
          qc->mv[qc->c] += 1 - gamma;
        #endif
        #endif
      }
    }
    #ifdef HALF_CV
    #if (HEUR_CONST & HEURISTIC_GV)
    success &= (qc->gv[qc->c] >= -EPSILON);
    #ifdef COUNT_HEUR
    if (qc->gv[qc->c] < -EPSILON)
      gvbound++;
    #endif

    #endif
    #if (HEUR_CONST & HEURISTIC_MV)
    success &= (qc->mv[qc->c] >= -EPSILON);
    #ifdef COUNT_HEUR
    if (qc->mv[qc->c] < -EPSILON)
      mvbound++;
    #endif

    #endif
    #endif
  }
  else
  {
    #ifdef DEBUG
    printf("Adding vertex %d FAILED!\n", vertex);
    fflush(stdout);
    #ifdef FULL_CV
    exit(1);
    #endif
    #endif
    G->vertex[qc->v[0]].active |= TO_REMOVE;
  }
  /*
    for (i = qc->c + 1; i < qc->n; i++)
      if (connected(G, qc->v[i], vertex) == 0)
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[i] -= gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_CV && defined(FULL_CV))
        if ((active & QUERY_VERTEX) == 0)
          qc->cv[i] -= gamma;
        #endif
      }
      else
      {
        #ifdef DEBUG
        qc->nc[i]++;
        #endif
        #if (HEUR_CONST & HEURISTIC_CV)
        #ifdef FULL_CV
        if ((active & QUERY_VERTEX) == 0)
          qc->cv[i] -= addadjnq;
        #endif
        #ifdef HALF_CV
        if ((G->vertex[qc->v[i]].active & QUERY_VERTEX) == 0)
          qc->cv[qc->c] += prepadjnqc;
        else
          qc->cv[qc->c] += 1 - mu;
        #endif
        #endif
      }
  //*/
  
  qc->c++;
  #ifdef DEBUG
  printf("%*sAdd(%d):  ", qc->c - 1, "", vertex);
  print_qc();
  verify_qc();
  fflush(stdout);
  #endif
  
  #if (TIME_CONST & TIME_ADD)
  resume_clock();
  #endif
  
  if (success == 0)
    for (i = qc->c; i < qc->n; i++)
      G->vertex[qc->v[i]].active &= ~TO_REMOVE;
  
  return success;
}

int check_candidates(int outer)
{
  int found, left, right;
  #if (TIME_CONST & TIME_CHECK)
  select_clock(&checktime);
  #endif
  #ifdef DEBUG
  printf("Checking candidate set...\n");
  fflush(stdout);
  #endif
  
  found = 0;
  right = qc->n;
  while (1)
  {
    for (left = qc->n - 1; left >= qc->c; left--)
      if (G->vertex[qc->v[left]].active & TO_REMOVE)
      {
        found = 1;
        //G->vertex[qc->v[left]].active &= ~TO_REMOVE;
        qc_swap(left, --right);
      }
    if (right == qc->n)
      break;
    
    if (remove_from_c(right, outer) == 0)
    {
      #ifdef DEBUG
      printf("Finished checking candidates\n");
      fflush(stdout);
      #endif
      return -1;
    }
  }
  
  #ifdef DEBUG
  printf("Done checking candidates\n");
  if (found)
    printf("(max check)\n");
  fflush(stdout);
  #endif
  
  #if (TIME_CONST & TIME_CHECK)
  select_clock(&enumtime);
  #endif
  return found;
}

void remove_from_qc()
{
  qc->c--;
  
  graph_enum_t vertex = qc->v[qc->c];
  int active = G->vertex[vertex].active;
  int i;
  
  if ((active & QUERY_VERTEX) == 0)
    #if (HEUR_CONST & HEURISTIC_E)
    qc->e += mu;
    #else
    ;
    #endif
  #ifdef DEBUG
  else
    qc->qvc--;
  #endif
  
  for (i = 0; i < qc->c; i++)
    if (connected(G, qc->v[i], vertex))
    {
      #if (HEUR_CONST & HEURISTIC_GV)
      if ((active & QUERY_VERTEX) == 0)
        qc->gv[i] += addgadjnq;
      #endif
      #ifdef DEBUG
      qc->nc[i]--;
      #endif
    }
    else
    {
      #if (HEUR_CONST & HEURISTIC_DV)
      qc->dv[i] += gamma;
      #endif
      #if (HEUR_CONST & HEURISTIC_MV)
      if ((active & QUERY_VERTEX) == 0)
      {
        qc->mv[i] += mu;
        #if (HEUR_CONST & HEURISTIC_GV)
        qc->gv[i] += gamma;
        #endif
      }
      else
        qc->mv[i] += addmnadjq;
      #elif (HEUR_CONST & HEURISTIC_GV)
      if ((active & QUERY_VERTEX) == 0)
        qc->gv[i] += gamma;
      #endif
    }
  if (G->vertex[qc->v[0]].active & TO_REMOVE)
    G->vertex[qc->v[0]].active &= ~TO_REMOVE;
  else
		for (i++; i < qc->n; i++)
			if (connected(G, qc->v[i], vertex) == 0)
			{
				#if (HEUR_CONST & HEURISTIC_DV)
				qc->dv[i] += gamma;
				#endif
				#ifdef FULL_CV
				#if (HEUR_CONST & HEURISTIC_MV)
			  if ((active & QUERY_VERTEX) == 0)
			  {
			    qc->mv[i] += mu;
			    #if (HEUR_CONST & HEURISTIC_GV)
			    qc->gv[i] += gamma;
			    #endif
			  }
        else
          qc->mv[i] += addmnadjq;
				#elif (HEUR_CONST & HEURISTIC_GV)
				if ((active & QUERY_VERTEX) == 0)
					qc->gv[i] += gamma;
				#endif
				#endif
			}
			else
			{
				#if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
				if ((active & QUERY_VERTEX) == 0)
					qc->gv[i] += addgadjnq;
				#endif
				#ifdef DEBUG
				qc->nc[i]--;
				#endif
			}
  
  #ifdef DEBUG
  printf("%*sRemove(%d):  ", qc->c, "", vertex);
  print_qc();
  verify_qc();
  fflush(stdout);
  #endif
}

int remove_from_c(int index, int outer)
{
  #if (TIME_CONST & TIME_REMOVE)
  pause_clock();
  select_clock(&remtime);
  #endif
  remcount++;
  
  #ifdef DEBUG
  if (index >= qc->n || index < qc->c)
  {
    printf("FAIL:  Toss(%d:%d)\n", qc->v[index], qc->v[qc->n - 1]);
    print_qc();
    exit(EXIT_FAILURE);
  }
  #endif
  
  int success = -1;
  int i, j;
  //qc->r = index;
  int last = qc->n;
  
  #if (HEUR_CONST & (HEURISTIC_CRIT_GV | HEURISTIC_CRIT_DV | HEURISTIC_CRIT_E))
  //do
  //{
  //index = qc->r;
  #ifdef DEBUG
  //if (last != qc->n)
  //  printf("(Back)\n");
  #endif
  #endif
  
  #if (HEUR_CONST & HEURISTIC_E)
  for (j = index; j < last; j++)
    if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
      qc->e -= 1 - mu;
  
  success &= (qc->e >= -EPSILON);
  #ifdef COUNT_HEUR
  if (qc->e < -EPSILON)
    ebound++;
  #endif
  #if (HEUR_CONST & HEURISTIC_CRIT_E)
  if ((qc->e >= -EPSILON) & (qc->e < mu - EPSILON))
  {
    #ifdef DEBUG
    printf("Mu is critical\n");
    fflush(stdout);
    #endif
    //for (j = qc->n - 1; j >= qc->c; j--)
    for (j = qc->c; j < qc->n; j++)
      if ((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0)
      {
        #ifdef DEBUG
        //printf("(Remove %d)\n", qc->v[j]);
        #endif
        //qc_swap(j, --qc->r);
        G->vertex[qc->v[j]].active |= TO_REMOVE;
        #ifdef COUNT_HEUR
        cecand++;
        #endif
      }
  }
  #endif
  #endif
  
  #ifdef DEBUG
  //printf("I:  (%d : %d)\n", qc->v[index], qc->v[last - 1]);
  #endif
  for (i = 0; i < qc->c; i++)
  {
    for (j = index; j < last; j++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[i] -= 1 - gamma;
        success &= (qc->dv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->dv[i] < -EPSILON)
          dvbound++;
        #endif
        #endif
        #if (HEUR_CONST & HEURISTIC_GV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
        {
          qc->gv[i] -= 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV)
          qc->mv[i] -= 1 - gamma;
          #endif
        }
        else
          qc->gv[i] -= remgadjnq;
        success &= (qc->gv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->gv[i] < -EPSILON)
          gvbound++;
        #endif
        #elif (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          qc->mv[i] -= 1 - gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_MV)
        success &= (qc->mv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->mv[i] < -EPSILON)
          mvbound++;
        #endif
        #endif
        #ifdef DEBUG
        //printf("%d~%d\n", qc->v[i], qc->v[j]);
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV)
      else if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
      {
        qc->mv[i] -= remmnadjq;
        success &= (qc->mv[i] >= -EPSILON);
        #ifdef COUNT_HEUR
        if (qc->mv[i] < -EPSILON)
          mvbound++;
        #endif

      }
    #if (HEUR_CONST & HEURISTIC_CRIT_DV)
    if (success & (qc->dv[i] < gamma - EPSILON))
    {
      #ifdef DEBUG
      printf("Vertex %d has a critical dv\n", qc->v[i]);
      fflush(stdout);
      #endif
      //for (j = qc->n - 1; j >= qc->c; j--)
      for (j = qc->c; j < qc->n; j++)
        if (connected(G, qc->v[i], qc->v[j]) == 0)
        {
          #ifdef DEBUG
          //printf(" (Remove %d)\n", qc->v[j]);
          //fflush(stdout);
          #endif
          //qc_swap(j, --qc->r);
          G->vertex[qc->v[j]].active |= TO_REMOVE;
          #ifdef COUNT_HEUR
          cdvcand++;
          #endif
        }
    }
    #endif
    #if (HEUR_CONST & HEURISTIC_CRIT_GV)
    if (success & (qc->gv[i] < gamma - EPSILON))
    {
      #ifdef DEBUG
      printf("Vertex %d has a critical gv (-N - Q)\n", qc->v[i]);
      fflush(stdout);
      #endif
      //for (j = qc->n - 1; j > qc->c; j--)
      for (j = qc->c + 1; j < qc->n; j++)
        if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]) == 0)
        {
          //qc_swap(j, --qc->r);
          G->vertex[qc->v[j]].active |= TO_REMOVE;
          #ifdef COUNT_HEUR
          cgvcand++;
          #endif
        }
    }
    if (success & (qc->gv[i] < addgadjnq - EPSILON))
    {
      #ifdef DEBUG
      printf("Vertex %d has a critical gv (N - Q)\n", qc->v[i]);
      fflush(stdout);
      #endif
      //for (j = qc->n - 1; j > qc->c; j--)
      for (j = qc->c + 1; j < qc->n; j++)
        if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]))
        {
          //qc_swap(j, --qc->r);
          G->vertex[qc->v[j]].active |= TO_REMOVE;
          #ifdef COUNT_HEUR
          cgvcand++;
          #endif
        }
    }
    #endif
    #if (HEUR_CONST & HEURISTIC_CRIT_MV)
    if (success & (qc->mv[i] < mu - EPSILON))
    {
      #ifdef DEBUG
      printf("Vertex %d has a critical mv (-N - Q)\n", qc->v[i]);
      fflush(stdout);
      #endif
      //for (j = qc->n - 1; j > qc->c; j--)
      for (j = qc->c + 1; j < qc->n; j++)
        if (((G->vertex[qc->v[j]].active & QUERY_VERTEX) == 0) && connected(G, qc->v[i], qc->v[j]) == 0)
        {
          //qc_swap(j, --qc->r);
          G->vertex[qc->v[j]].active |= TO_REMOVE;
          #ifdef COUNT_HEUR
          cmvcand++;
          #endif
        }
    }
    if (success & (qc->mv[i] < addmnadjq - EPSILON))
    {
      #ifdef DEBUG
      printf("Vertex %d has a critical mv (Q - N)\n", qc->v[i]);
      fflush(stdout);
      #endif
      //for (j = qc->n - 1; j > qc->c; j--)
      for (j = qc->c + 1; j < qc->n; j++)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX && (connected(G, qc->v[i], qc->v[j]) == 0))
        {
          //qc_swap(j, --qc->r);
          G->vertex[qc->v[j]].active |= TO_REMOVE;
          #ifdef COUNT_HEUR
          cmvcand++;
          #endif
        }
    }
    #endif
    #endif
  }
  
  #if (HEUR_CONST & (HEURISTIC_CRIT_CV | HEURISTIC_CRIT_DV | HEURISTIC_CRIT_E))
  //last = index;
  //} while (success & (qc->r < index));
  #endif
  
  #ifdef DEBUG
  //printf("Comparing (%d : %d) to (%d : %d)\n", qc->v[qc->c], qc->v[index - 1], qc->v[index], qc->v[qc->n - 1]);
  //printf("Intermediate (%d): ", success);
  //print_qc();
  //fflush(stdout);
  #endif
  
  //for (i = index - 1; i >= qc->c; i--)
  for (j = index; j < qc->n; j++)
    for (i = qc->c; i < index; i++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[i] -= 1 - gamma;
        #endif
        #ifdef FULL_CV
        #if (HEUR_CONST & HEURISTIC_GV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
        {
          qc->gv[i] -= 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV)
          qc->mv[i] -= 1 - gamma;
          #endif
        }
        else
          qc->gv[i] -= remgadjnq;
        #elif (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          qc->mv[i] -= 1 - gamma;
        #endif
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
      else if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
        qc->mv[i] -= remmnadjq;
      #endif
  
  if (success)
    for (i = qc->c; i < index; i++)
    {
      #if (HEUR_CONST & HEURISTIC_DV)
      if (qc->dv[i] < -EPSILON)
      {
        #ifdef DEBUG
        printf("Vertex %d has dv < 0\n", qc->v[i]);
        fflush(stdout);
        #endif
        G->vertex[qc->v[i]].active |= TO_REMOVE;
        #ifdef COUNT_HEUR
        dvcand++;
        #endif
        //qc_swap(i, --qc->r);
      }
      #if (HEUR_CONST & (HEURISTIC_GV | HEURISTIC_MV) && defined(FULL_CV))
      else
      #endif
      #endif
      #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
      if (qc->gv[i] < -EPSILON)
      {
        #ifdef DEBUG
        printf("Vertex %d has gv < 0\n", qc->v[i]);
        fflush(stdout);
        #endif
        G->vertex[qc->v[i]].active |= TO_REMOVE;
        #ifdef COUNT_HEUR
        gvcand++;
        #endif
        //qc_swap(i, --qc->r);
      }
      #if (HEUR_CONST & HEURISTIC_MV)
      else
      #endif
      #endif
      #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
      if (qc->mv[i] < -EPSILON)
      {
        #ifdef DEBUG
        printf("Vertex %d has mv < 0\n", qc->v[i]);
        fflush(stdout);
        #endif
        G->vertex[qc->v[i]].active |= TO_REMOVE;
        #ifdef COUNT_HEUR
        mvcand++;
        #endif
        //qc_swap(i, --qc->r);
      }
      #endif
    }
  
  for (i = qc->n; i < outer; i++)
    for (j = index; j < qc->n; j++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[j] += 1 - gamma;
        #endif
        #ifdef FULL_CV
        #if (HEUR_CONST & HEURISTIC_GV)
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        {
          qc->gv[j] += 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV)
          qc->mv[j] += 1 - gamma;
          #endif
        }
        else
          qc->gv[j] += remgadjnq;
        #elif (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
          qc->mv[j] += 1 - gamma;
        #endif
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
      else if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        qc->mv[j] += remmnadjq;
      #endif
  
  #ifdef DEBUG
  printf("%*sToss(%d:%d):  ", qc->c, "", qc->v[index], qc->v[qc->n - 1]);
  fflush(stdout);
  #endif
  
  qc->n = index;
  
  #ifdef DEBUG
  print_qc();
  verify_qc();
  #endif
  
  if (success == 0)
    for (i = qc->c; i < qc->n; i++)
      G->vertex[qc->v[i]].active &= ~TO_REMOVE;
  
  #if (TIME_CONST & TIME_REMOVE)
  resume_clock();
  #endif
  
  return success;
}

void return_to_c(int index)
{
  int i, j;
  #ifdef DEBUG
  printf("Return_to_c(%d)\n", index);
  fflush(stdout);
  #endif
  
  #if (HEUR_CONST & HEURISTIC_E)
  for (j = qc->n; j < index; j++)
  {
    if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
      qc->e += 1 - mu;
    G->vertex[qc->v[j]].active &= ~TO_REMOVE;
  }
  #endif
  
  for (i = qc->n; i < index; i++)
  {
    for (j = 0; j < qc->c; j++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[j] += 1 - gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_GV)
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        {
          qc->gv[j] += 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV)
          qc->mv[j] += 1 - gamma;
          #endif
        }
        else
          qc->gv[j] += remgadjnq;
        #elif (HEUR_CONST & HEURISTIC_MV)
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
          qc->mv[j] += 1 - gamma;
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV)
      else if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        qc->mv[j] += remmnadjq;
      #endif
    
    for (; j < qc->n; j++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        #if (HEUR_CONST & HEURISTIC_DV)
        qc->dv[j] += 1 - gamma;
        #endif
        #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        {
          qc->gv[j] += 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV)
          qc->mv[j] += 1 - gamma;
          #endif
        }
        else
          qc->gv[j] += remgadjnq;
        #elif (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
          qc->mv[j] += 1 - gamma;
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
      else if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        qc->mv[j] += remmnadjq;
      #endif
  }
  
  #ifdef DEBUG
  printf("%*sReturn(%d:%d)\n", qc->c, "", qc->v[qc->n], qc->v[index - 1]);
  fflush(stdout);
  //print_qc();
  #endif
  
  qc->r = qc->n = index;
  
  #ifdef DEBUG
  verify_qc();
  #endif
}

void prepare_qc(int vertex)
{
  #if (TIME_CONST & TIME_PREP)
  pause_clock();
  select_clock(&preptime);
  #endif
  
  qc->v[0] = vertex;
  G->vertex[vertex].active |= QUERIED;
  qc->s = 0;
  qc->c = 1;
  
  #if (HEUR_CONST & HEURISTIC_GV)
  qc->gv[0] = gamma * (1 - mu);
  #endif
  #if (HEUR_CONST & HEURISTIC_MV)
  qc->mv[0] = (1 - mu) * (1 - gamma);
  #endif
  #if (HEUR_CONST & HEURISTIC_DV)
  qc->dv[0] = 0;
  #endif
  #if (HEUR_CONST & HEURISTIC_E)
  qc->e = 1 - mu;
  #endif
  #ifdef DEBUG
  qc->nc[0] = 0;
  qc->qvc = 1;
  #endif
  
  #ifdef DEBUG
  printf("Vertex %d (%d) has %d vertices within distance 2\n", vertex, G->vertex[vertex].degree, G->vertex[vertex].n2_size + G->vertex[vertex].degree);
  fflush(stdout);
  #endif
  
  int i, j;
  qc->n = 1;
  for (i = 0; i < G->vertex[vertex].degree; i++)
  {
    graph_enum_t neighbor = G->vertex[vertex].edge_list[i].vertex;
    if ((G->vertex[neighbor].active & QUERIED) == 0)
      qc->v[qc->n++] = neighbor;
  }
  for (i = 0; i < G->vertex[vertex].n2_size; i++)
  {
    graph_enum_t n2 = G->vertex[vertex].n2_list[i];
    if ((G->vertex[n2].active & QUERIED) == 0)
      qc->v[qc->n++] = n2;
  }
  
  //G->vertex[vertex].active &= ~IN_QC;
  for (i = qc->n - 1; i > 0; i--)
  {
    if (connected(G, vertex, qc->v[i]))
    {
      if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
      {
        #if (HEUR_CONST & HEURISTIC_GV)
        qc->gv[0] += 1 - mu;
        #ifdef FULL_CV
        qc->gv[i] = (gamma + 1) * (1 - mu);
        #endif
        #endif
        #if (HEUR_CONST & HEURISTIC_MV)
        qc->mv[0] += 1 - gamma;
        #ifdef FULL_CV
        qc->mv[i] = (2 - mu) * (1 - gamma);
        #endif
        #endif
        #if (HEUR_CONST & HEURISTIC_E)
        qc->e += 1 - mu;
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_GV)
      else
      {
        qc->gv[0] += prepgadjnqc;
        #ifdef FULL_CV
        qc->gv[i] = (gamma + 1) * (1 - mu) - gamma;
        #if (HEUR_CONST & HEURISTIC_MV)
        qc->mv[i] = (1 - mu) * (1 - gamma);
        #endif
        #endif
      }
      #elif (HEUR_CONST & HEURISTIC_MV)
      else
        qc->mv[i] = (1 - mu) * (1 - gamma);
      #endif
      
      #if (HEUR_CONST & HEURISTIC_DV)
      qc->dv[0] += 1 - gamma;
      qc->dv[i] = 1 - gamma;
      #ifdef DEBUG
      printf("qc->dv[%d] = %lf\n", qc->v[i], 1 - gamma);
      fflush(stdout);
      #endif
      #endif
      #ifdef DEBUG
      qc->nc[i] = 1;
      #endif
    }
    else
    {
      #if (HEUR_CONST & HEURISTIC_DV)
      if (G->vertex[qc->v[i]].degree < gamma / (1 - gamma) - EPSILON)
      {
        #ifdef DEBUG
        printf("Vertex %d being removed\n", qc->v[i]);
        fflush(stdout);
        #endif
        #ifdef COUNT_HEUR
        lowdegcand++;
        #endif
        qc_swap(i, --qc->n);
        continue;
      }
      qc->dv[i] = -gamma;
      #ifdef DEBUG
      printf("qc->dv[%d] = %lf\n", qc->v[i], -gamma);
      fflush(stdout);
      #endif
      #endif
      #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
      if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
      {
        qc->gv[i] = gamma * (1 - mu);
        #if (HEUR_CONST & HEURISTIC_MV)
        qc->mv[0] += prepmnadjqc;
        qc->mv[i] = (2 - mu) * (1 - gamma) - mu;
        #endif
      }
      else
      {
        qc->gv[i] = -gamma * mu;
        #if (HEUR_CONST & HEURISTIC_MV)
        qc->mv[i] = 1 - gamma - mu * (2 - gamma);
        #endif
      }
      #elif (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
      if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
      {
        qc->mv[0] += prepmnadjqc;
        qc->mv[i] = (1 - mu) * (1 - gamma) - mu;
      }
      else
        qc->mv[i] = mu * (gamma - 2);
      #elif (HEUR_CONST & HEURISTIC_MV)
      if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        qc->mv[0] += prepmnadjqc;
      #endif
      
      #if (HEUR_CONST & HEURISTIC_E)
      if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        qc->e += 1 - mu;
      #endif
      #ifdef DEBUG
      qc->nc[i] = 0;
      #endif
    }
    
    #if ((HEUR_CONST & HEURISTIC_DV) || (defined(FULL_CV) && (HEUR_CONST & (HEURISTIC_GV | HEURISTIC_MV))))
    for (j = i + 1; j < qc->n; j++)
      if (connected(G, qc->v[i], qc->v[j]))
      {
        #ifdef DEBUG
        printf("(%d) qc->dv[%d] : %lf -> ", qc->v[i], qc->v[j], qc->dv[j]);
        fflush(stdout);
        #endif
        #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
        {
          qc->gv[i] += 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
          qc->mv[i] += 1 - gamma;
          #endif
        }
        else
          qc->gv[i] += prepgadjnqc;
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        {
          qc->gv[j] += 1 - mu;
          #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
          qc->mv[j] += 1 - gamma;
          #endif
        }
        else
          qc->gv[j] += prepgadjnqc;
        #elif (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
          qc->mv[i] += 1 - gamma;
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
          qc->mv[j] += 1 - gamma;
        #endif
        qc->dv[i] += 1 - gamma;
        qc->dv[j] += 1 - gamma;
        #ifdef DEBUG
        printf("%lf\n", qc->dv[j]);
        fflush(stdout);
        #endif
      }
      #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
      else
      {
        if (G->vertex[qc->v[j]].active & QUERY_VERTEX)
        {
          qc->mv[i] += prepmnadjqc;
        }
        if (G->vertex[qc->v[i]].active & QUERY_VERTEX)
        {
          qc->mv[j] += prepmnadjqc;
        }
      }
      #endif
    #endif
  }
  qc->size = qc->r = qc->n;
  
  #if ((HEUR_CONST & HEURISTIC_DV) || (defined(FULL_CV) && (HEUR_CONST & (HEURISTIC_GV | HEURISTIC_MV))))
  //for (i = qc->size - 1; i > 0; i--)
  for (i = qc->c + 1; i < qc->size; i++)
  {
    #if (HEUR_CONST & HEURISTIC_DV)
    if (qc->dv[i] < -EPSILON)
    {
      //qc_swap(i, --qc->r);
      G->vertex[qc->v[i]].active |= TO_REMOVE;
      #ifdef COUNT_HEUR
      dvcand++;
      #endif
    }
    #endif
    #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
    if (qc->gv[i] < -EPSILON)
    {
      //qc_swap(i, --qc->r);
      G->vertex[qc->v[i]].active |= TO_REMOVE;
      #ifdef COUNT_HEUR
      gvcand++;
      #endif
    }
    #endif
    #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
    if (qc->mv[i] < -EPSILON)
    {
      //qc_swap(i, --qc->r);
      G->vertex[qc->v[i]].active |= TO_REMOVE;
      #ifdef COUNT_HEUR
      mvcand++;
      #endif
    }
    #endif
  }
  #endif
  
  #ifdef DEBUG
  printf("Prepare: ");
  print_qc();
  fflush(stdout);
  #endif
  
  #if (TIME_CONST & TIME_PREP)
  resume_clock();
  #endif
}

void output_qc()
{
  #if (TIME_CONST & TIME_OUTPUT)
  pause_clock();
  select_clock(&outtime);
  #endif
  
  #ifdef DEBUG
  if (is_qc())
    printf("CLIQUE %d: ", num_qcs);
  else
    printf("CLIQUE %d (FAILED): ", num_qcs);
  print_qc();
  fflush(stdout);
  #endif
  
  int i = qc->s;
  
  fprintf(out, "%d", qc->v[i]);
  for (i++; i < qc->c; i++)
    fprintf(out, " %d", qc->v[i]);
  fprintf(out, "\n");
  #ifdef FLUSH_OUT
  fflush(out);
  #endif
  
  #if (HEUR_CONST & HEURISTIC_MAX)
  #ifdef HBI
  if (num_qcs == delta[maxlevel + 1] - delta[maxlevel])
  {
    maxlevel++;
    //printf("RESIZING mbytes=%lu: num_qcs == delta[maxlevel + 1] - delta[maxlevel] == %d\n", (delta[maxlevel] * G->num_vertices * sizeof(index_t))/1048576, num_qcs);
    
    #ifdef DEBUG
    printf("Reallocating maximality index!\n");
    printf("Old delta: ");
    for (i = 0; i <= maxlevel; i++)
      printf(" %d", delta[i]);
    fflush(stdout);
    #endif
    //Reallocate
    char* temp = (char*) malloc((maxlevel + 2) * sizeof(unsigned long));
    if (temp == NULL)
    {
      perror("Error enlarging maximality index helper array");
      exit(EXIT_FAILURE);
    }
    memcpy(temp, delta, (maxlevel + 1) * sizeof(unsigned long));
    free(delta);
    delta = (unsigned long*) temp;
    delta[maxlevel + 1] = delta[maxlevel] * BRANCH_FACTOR + 1;
    #ifdef DEBUG
    printf("\nNew delta: ");
    for (i = 0; i <= maxlevel; i++)
      printf(" %d", delta[i]);
    printf(" %d\n", delta[maxlevel + 1]);
    fflush(stdout);
    #endif

    temp = (char*) malloc(delta[maxlevel] * G->num_vertices * sizeof(index_t));
    if (temp == NULL)
    {
      perror("Error enlarging quasi-clique maximality index");
      exit(EXIT_FAILURE);
    }

    memcpy(temp, qc_index, delta[maxlevel-1] * (G->num_vertices) * sizeof(index_t));    //maxlevel-1 to because thats how big qc_index is
    memset(temp + delta[maxlevel - 1] * G->num_vertices, 0, (delta[maxlevel] - delta[maxlevel - 1]) * (G->num_vertices * sizeof(index_t)));
    #ifdef DEBUG
    printf("New memory zeroed\n");
    fflush(stdout);
    #endif
    int newidx = max_index(maxlevel, 0) * G->num_vertices;
    int oldidx = max_index(maxlevel - 1, 0) * G->num_vertices;
    for (i = 0; i < G->num_vertices; i++)
      temp[newidx + i] = (temp[oldidx + i] != 0);
    free(qc_index);
    qc_index = (index_t*) temp;
  }
  #else 
  if (num_qcs == BRANCH_FACTOR * index_len)
  {
      print("HERE!\n");
    #ifdef DEBUG
    printf("Reallocating maximality index!\n");
    fflush(stdout);
    #endif
    
    index_t* temp = (index_t*) malloc((index_len * G->num_vertices * sizeof(index_t)) * 2);
    if (temp == NULL)
    {
      perror("Error enlarging quasi-clique maximality index");
      exit(EXIT_FAILURE);
    }
    memcpy(temp, qc_index, index_len * G->num_vertices * sizeof(index_t));
    #ifdef DEBUG
    printf("Old index copied\n");
    fflush(stdout);
    #endif
    memset(temp + index_len * G->num_vertices, 0, index_len * G->num_vertices * sizeof(index_t));
    #ifdef DEBUG
    printf("New memory zeroed\n");
    fflush(stdout);
    #endif
    free(qc_index);
    qc_index = temp;
    index_len <<= 1;
  }
  #endif
  
  int word, pos, level, idx;
  word = num_qcs / BRANCH_FACTOR;
  pos  = num_qcs % BRANCH_FACTOR;
  #ifdef HBI
  for (level = 0; level <= maxlevel; level++)
  {
  idx = max_index(level, word);
  #else
  #ifdef DEBUG
  
  if (word >= index_len)
  {
    fprintf(stderr, "word = %d >= %d = index_len\n", word, index_len);
    exit(EXIT_FAILURE);
  }
  #endif
  idx = word;
  #endif
  for (i = 0; i < qc->c; i++)
  {
    #ifdef DEBUG
    #ifdef HBI
    printf("qc_index[%d][%d, %d]:  %d -> ", qc->v[i], maxlevel - level, word, qc_index[idx * G->num_vertices + qc->v[i]]);
    fflush(stdout);
    #endif
    #endif
    qc_index[(unsigned long)idx * G->num_vertices + qc->v[i]] |= (1 << pos);
    #ifdef DEBUG
    #ifdef HBI
    printf("%d\n", qc_index[idx * G->num_vertices + qc->v[i]]);
    fflush(stdout);
    #endif
    #endif
  }
  
  #ifdef HBI
  pos  = word % BRANCH_FACTOR;
  word = word / BRANCH_FACTOR;
  }
  #endif
  #endif
  
  num_qcs++;
  //if (num_qcs%10000 == 0)
  //    printf("num_qcs = %d\n", num_qcs);
  
  #if (TIME_CONST & TIME_OUTPUT)
  resume_clock();
  #endif
}

#if (HEUR_CONST & HEURISTIC_MAX)
#ifdef HBI
int max_index(int level, int word)
{
  #ifdef DEBUG
  printf("Getting index\n");
  fflush(stdout);
  #endif
  int block = (word + 1) * delta[level] - 1;
  while (word >= BRANCH_FACTOR)
  {
    word /= BRANCH_FACTOR;
    block += word;
  }
  
  #ifdef DEBUG
  if (block >= delta[maxlevel])
  {
    fprintf(stderr, "Tried to access block %d (%d, %d) out of %d\n", block, level, word, delta[maxlevel]);
    exit(EXIT_FAILURE);
  }
  else
  {
    //printf("Block = %d\n", block);
    //fflush(stdout);
  }
  #endif
  
  return block;
}

/*
void update_max(int level, int word)
{
  int idx = max_index(level, word);
  tempindex[idx] = currindex[idx] &= qc_index[idx * G->num_vertices + qc->v[qc->c - 1]];

  if (level > 0 && currindex[idx])
  {
    index_t mask;
    int t;
    for (mask = 1, t = word * BRANCH_FACTOR; mask != 0; mask <<= 1, t++)
      if (currindex[idx] & mask)
        update_max(level - 1, t);
  }
}
//*/

index_t check_max(int level, int word)
{
  int idx = max_index(level, word);
  #ifdef DEBUG
  printf("%*sChecking byte[%d, %d]...\n", maxlevel - level + 1, "", maxlevel - level, word);
  fflush(stdout);
  #endif
  int i;
  index_t temp;
    
    //printf("idx=%d, num_v=%d, +qc_v=%d\n",idx, G->num_vertices, qc->v[qc->c-1]);
  temp = qc_index[(unsigned long)idx * G->num_vertices + qc->v[qc->c - 1]];
  #ifdef DEBUG
  printf("%*sqc_index[%d][%d, %d] = %d\n", maxlevel - level + 1, "", qc->v[qc->c - 1], maxlevel - level, word, qc_index[idx * G->num_vertices + qc->v[qc->c - 1]]);
  fflush(stdout);
  #endif
  for (i = 0; (i < qc->n) & (temp != 0); i++)
  {
    #ifdef DEBUG
    printf("%*sqc_index[%d][%d, %d] = %d\n", maxlevel - level + 1, "", qc->v[i], maxlevel - level, word, qc_index[idx * G->num_vertices + qc->v[i]]);
    fflush(stdout);
    #endif
    temp &= qc_index[(unsigned long)idx * G->num_vertices + qc->v[i]];
  }
  
  if (level == 0)
  {
    #ifdef DEBUG
    printf("%*sByte[%d, %d] %s\n", maxlevel - level + 1, "", maxlevel - level, word, temp? "not cleared":"cleared");
    fflush(stdout);
    #endif
    return temp == 0;
  }
  
  #ifdef DEBUG
  printf("Checking temp = %d\n", temp);
  fflush(stdout);
  #endif
  if (temp != 0)
  {
    index_t mask;
    for (mask = 1, i = word * BRANCH_FACTOR; mask != 0; mask <<= 1, i++)
    {
      if ((temp & mask) != 0 && check_max(level - 1, i) == 0)
      {
        #ifdef DEBUG
        printf("%*sFailed on bit %d!\n", maxlevel - level + 1, "", i - word * BRANCH_FACTOR);
        fflush(stdout);
        #endif
        return 0;
      }
    }
  }
  
  #ifdef DEBUG
  printf("%*sByte[%d, %d] clear!\n", maxlevel - level + 1, "", maxlevel - level, word);
  fflush(stdout);
  #endif
  return 1;
}
#endif

int check_maximality()
{
  #if (TIME_CONST & TIME_MAX)
  pause_clock();
  select_clock(&maxtime);
  #endif
  
  #ifdef DEBUG
  printf("Checking maximality\n");
  fflush(stdout);
  //print_qc();
  #endif
  //update_max(maxlevel, 0);
  
  #if (HEUR_CONST & HEURISTIC_MIN_QC_SIZE)
  if (qc->n < min_qc_size)
  {
    #if (TIME_CONST & TIME_MAX)
    resume_clock();
    #endif
    return 0;
  }
  #endif
  
  #ifdef HBI
  index_t retval = check_max(maxlevel, 0);
  #else
  if (num_qcs == 0)
    return 1;
  
  #ifdef DEBUG
  printf("%d\n", (num_qcs - 1) / BRANCH_FACTOR);
  fflush(stdout);
  if ((num_qcs - 1) / BRANCH_FACTOR >= index_len)
  {
    fprintf(stderr, "index_len = %d\n", index_len);
    exit(EXIT_FAILURE);
  }
  #endif
  index_t retval = 0;
  int i, word;
  for (word = 0; (word <= (num_qcs - 1) / BRANCH_FACTOR) & (retval == 0); word++)
  {
    retval = qc_index[word * G->num_vertices + qc->v[qc->c - 1]];
    for (i = 0; (i < qc->n) & (retval != 0); i++)
      retval &= qc_index[word * G->num_vertices + qc->v[i]];
    #ifdef DEBUG
    printf("Word %d / %d %s (0x%x)!\n", word, (num_qcs - 1) / BRANCH_FACTOR, retval? "not clear":"clear", retval);
    fflush(stdout);
    #endif
  }
  retval = (retval == 0);
  #endif
  
  #if (TIME_CONST & TIME_MAX)
  resume_clock();
  #endif
  #ifdef DEBUG
  printf("Maximal? %d\n", retval);
  fflush(stdout);
  #endif
  return retval;
}
#endif

void qc_swap(int index1, int index2)
{
  if (index1 != index2)
  {
    graph_enum_t temp = qc->v[index2];
    qc->v[index2] = qc->v[index1];
    qc->v[index1] = temp;
    
    double temp_d;
    #if (HEUR_CONST & HEURISTIC_DV)
    temp_d = qc->dv[index2];
    qc->dv[index2] = qc->dv[index1];
    qc->dv[index1] = temp_d;
    #endif
    
    #ifdef DEBUG
    temp = qc->nc[index2];
    qc->nc[index2] = qc->nc[index1];
    qc->nc[index1] = temp;
    #endif
    
    #if (HEUR_CONST & HEURISTIC_GV && defined(FULL_CV))
    temp_d = qc->gv[index2];
    qc->gv[index2] = qc->gv[index1];
    qc->gv[index1] = temp_d;
    #endif
    #if (HEUR_CONST & HEURISTIC_MV && defined(FULL_CV))
    temp_d = qc->mv[index2];
    qc->mv[index2] = qc->mv[index1];
    qc->mv[index1] = temp_d;
    #endif
  }
}

int enumerate()
{
  enumcount++;
  
  #ifdef DEBUG
  //printf("in_enum:\n");
  //print_qc();
  #endif
  
  int outer = qc->n;
  int subsumed = 0;
  
  int candidates = check_candidates(outer);
  
  if (candidates == -1)
  {
    #ifdef DEBUG
    printf("Remove failed!\n");
    fflush(stdout);
    #endif
    return_to_c(outer);
    return 0;
  }
  #if (HEUR_CONST & HEURISTIC_MAX)
  else if (candidates != 0 && check_maximality() == 0)
  {
    #ifdef DEBUG
    printf("Maximality check failed!\n");
    fflush(stdout);
    #endif
    return_to_c(outer);
    return subsumed;
  }
  #endif
  
  /*
  if (qc->r < qc->n)
  {
    do
    {
      if (remove_from_c(qc->r, outer) == 0)
      {
        #ifdef DEBUG
        printf("Done checking candidates\n(remove failed)\n");
        fflush(stdout);
        #endif
        return_to_c(outer);
        return subsumed;
      }
    } while (qc->r < qc->n);
    
    #if (HEUR_CONST & HEURISTIC_MAX)
    if (check_maximality() == 0)
    {
      #ifdef DEBUG
      printf("Maximality check failed!\n");
      fflush(stdout);
      #endif
      subsumed = qc->n;
      return_to_c(outer);
      return subsumed;
    }
    #endif
  }
  //*/
  
  /*
  if (qc->n == qc->c)
  {
    if (qc->n < outer)
      return_to_c(outer);
    if (is_max_mgqc())
    {
      output_qc();
      return qc->n;
    }
    else
      return 0;
  }
  //*/
  
  while (qc->n > qc->c)
  {
    candidate_vertex();
    if (add_to_qc())
      subsumed = enumerate();
    remove_from_qc();
    
    //Backtrack until a return_to_c() call
    if (subsumed > 0)
      if (outer <= subsumed)
      {
        #ifdef DEBUG
        printf("Subsumed! (%d)\n", subsumed);
        fflush(stdout);
        #endif
        return subsumed;
      }
      else if (qc->n <= subsumed)
      {
        #ifdef DEBUG
        printf("Subsumed (%d -> %d)\n", subsumed, outer);
        fflush(stdout);
        #endif
        if (qc->n < outer)
          return_to_c(outer);
        return -1;
      }
      else
      {
        #ifdef DEBUG
        printf("Subsumption managed (%d > %d)\n", qc->n, subsumed);
        fflush(stdout);
        #endif
        subsumed = -1;
      }
    
    qc_swap(qc->c, qc->n - 1);
    
    if (remove_from_c(qc->n - 1, outer) == 0)
    {
      #ifdef DEBUG
      printf("Remove failed\n");
      fflush(stdout);
      #endif
      return_to_c(outer);
      return subsumed;
    }
    candidates = check_candidates(outer);
    if (candidates == -1)
    {
      #ifdef DEBUG
      printf("Remove failed.\n");
      fflush(stdout);
      #endif
      if (qc->n < outer)
        return_to_c(outer);
      return subsumed;
    }
    #if (HEUR_CONST & HEURISTIC_MAX)
    else if (check_maximality() == 0)
    {
      #ifdef DEBUG
      printf("Maximality check failed!\n");
      fflush(stdout);
      #endif
      subsumed = qc->n;
      if (qc->n < outer)
        return_to_c(outer);
      return subsumed;
    }
    #endif
    /*
    if (qc->r < qc->n)
      do
      {
        if (remove_from_c(qc->r, outer) == 0)
        {
          #ifdef DEBUG
          printf("Done checking candidates\n(remove failed)\n");
          fflush(stdout);
          #endif
          qc->r = outer;
          return_to_c(outer);
          return subsumed;
        }
      } while (qc->r < qc->n);
    
    #if (HEUR_CONST & HEURISTIC_MAX)
    if (check_maximality() == 0)
    {
      #ifdef DEBUG
      printf("Maximality check failed!\n");
      fflush(stdout);
      #endif
      subsumed = qc->n;
      return_to_c(outer);
      return subsumed;
    }
    #endif
    //*/
  }
  
  #ifdef DEBUG
  printf("out_enum %s: ", subsumed? "(subsumed)":"");
  int i;
  for (i = 0; i < qc->c; i++)
    printf(" %d", qc->v[i]);
  printf(" |");
  for (; i < qc->n; i++)
    printf(" %d", qc->v[i]);
  printf(" |");
  for (; i < qc->size; i++)
    printf(" %d", qc->v[i]);
  printf("\n");
  fflush(stdout);
  #endif
  
  if (subsumed == 0)
  {
    //printf("maxlevel = %d\n", maxlevel);
    output_qc();
    subsumed = qc->n;
  }
  
  if (qc->n < outer)
    return_to_c(outer);
  
  return subsumed;
}
