#ifndef __VERTEX_SET
#define __VERTEX_SET

#include <stdio.h>
#include <stdlib.h>

#include "vertexset1c.h"

//#define DEBUG

void printvs(PVertexSet set)
{
  PVertexSet end = set->p;
  PVertexSet curr = set + 1;
  while (curr < end)
  {
    printf("%d ", curr->v);
    curr++;
  }
  printf("\n");
}

void init()
{
  current_block = (PVertexSet) malloc(sizeof(PVertex) * VERTEX_SET_BLOCK_SIZE);
  if (current_block == NULL)
  {
    perror("Error allocating memory for vertex sets: ");
    exit(1);
  }
  current_block[0].p = NULL;
  current_block[1].p = current_block[2].p = current_block + 2;
}

void cleanup()
{
  #ifdef DEBUG
  printf("Deallocating vertex sets...\n");
  #endif
  
  PVertexSet next_block;
  while (current_block != NULL)
  {
    next_block = current_block->p;
    free(current_block);
    current_block = next_block;
  }
}

PGraph* newGraph(int num_vertices, int num_edges)
{
  PGraph* ret;
  int i;
  
  ret = (PGraph*) malloc(sizeof(PGraph));
  if (ret == NULL)
  {
    perror("Error allocating graph: ");
    exit(1);
  }
  
  ret->num_vertices = num_vertices;
  ret->num_edges = num_edges;

  ret->neighborhood_block = (PVertexSet) malloc(sizeof(PVertex)
                            * (2 + num_vertices + num_edges + num_edges));
  if (ret->neighborhood_block == NULL)
  {
    perror("Error allocating memory for storing vertex neighborhoods: ");
    exit(1);
  }
  ret->neighborhood_block[0].p = ret->neighborhood_block[1].p = 
  ret->neighborhood_block[2].p = ret->neighborhood_block + 2;
  
  ret->neighborhood = (PVertexSet*) malloc(sizeof(PVertexSet) * num_vertices);
  if (ret->neighborhood == NULL)
  {
    perror("Error allocating neighborhood sets: ");
    exit(1);
  }
  
  for (i = 0; i < num_vertices; i++)
    ret->neighborhood[i] = ret->neighborhood_block + 1;
  
  ret->marked = (char*) calloc(num_vertices, 1);
  if (ret->marked == NULL)
  {
    perror("Error allocating memory vertex marks: ");
    exit(1);
  }

  return ret;
}

void addEdge(int u, int v, PGraph* G)
{
  PVertexSet n = G->neighborhood_block->p;
  if (G->neighborhood[u] == n)
  {
    n->p++;
    n = n->p - 1;
    
    n->v = v;
  }
  else
  {
    *G->neighborhood_block = *n;
    n = G->neighborhood[u] = n->p;
    n->p = n + 2;
    n[1].v = v;
  }
  
  #ifdef DEBUG
  /*
  n = G->neighborhood_block;
  PVertexSet end = n->p;
  end = end->p;
  printf("%lu(%lu): ", G->neighborhood_block, end);
  n--;
  while (++n < end)
    printf("%lu ", n->p > 0 ?n->p :n->v);
  printf("\n");
  //*/
  #endif
}

void freeGraph(PGraph* G)
{
  #ifdef DEBUG
  printf("Deallocating graph...\n");
  #endif
  free(G->neighborhood_block);
  free(G->neighborhood);
  free(G->marked);
  free(G);
}

void allocateNewBlock()
{
  #ifdef DEBUG
  printf("ALLOCATING NEW BLOCK\n");
  #endif
  PVertexSet newblock = (PVertexSet) malloc(sizeof(PVertex) * VERTEX_SET_BLOCK_SIZE);
  if (newblock == NULL)
  {
    perror("Error allocating memory for new vertex sets: ");
    exit(1);
  }
  newblock[0].p = current_block;
  newblock[1].p = newblock + 2;
  current_block = newblock;
  vs_frame_pointer = current_block[1].p;
}

PVertexSet newVertexSet()
{
  PVertexSet ret = current_block[1].p;
  ret = ret->p;

  #ifdef DEBUG
  printf("New vertex set (%d)\n", ret + 2 - current_block);
  #endif
  
  if (ret + 2 - current_block >= VERTEX_SET_BLOCK_SIZE)
  {
    allocateNewBlock();
    ret = current_block + 2;
  }
  
  ret->p = ret + 1;
  current_block[1].p = ret;
  return ret;
}

PVertexSet setCopy(PVertexSet a)
{
  #ifdef DEBUG
  printf("Copy: ");
  printvs(a);
  #endif
  
  PVertexSet ret = current_block[1].p;
  ret = ret->p;
  
  int size = getSize(a) + 1;
  
  if (ret + size - current_block >= VERTEX_SET_BLOCK_SIZE)
  {
    allocateNewBlock();
    ret = current_block + 2;
  }
  
  long diff = ret - a;
  PVertexSet cpy;
  for (cpy = a + 1; cpy < a->p; cpy++)
    *(cpy + diff) = *cpy;
  
  ret->p = ret + size;
  current_block[1].p = ret;
  #ifdef DEBUG
  printf("After: ");
  printvs(ret);
  #endif
  return ret;
}

PVertexSet deepCopy(PVertexSet a)
{
  #ifdef DEBUG
  printf("Deep copy: ");
  printvs(a);
  #endif
  
  int size = getSize(a) + 1;
  
  PVertexSet ret = (PVertexSet) malloc(size * sizeof(PVertex));
  if (ret == NULL)
  {
    perror("Error allocating vertex set copy");
    exit(EXIT_FAILURE);
  }
  
  long diff = ret - a;
  PVertexSet cpy;
  for (cpy = a + 1; cpy < a->p; cpy++)
    *(cpy + diff) = *cpy;
  
  ret->p = ret + size;
  #ifdef DEBUG
  printf("After: ");
  printvs(ret);
  #endif
  return ret;
}

PVertexSet addVertex(PVertex u, PVertexSet a)
{
  #ifdef DEBUG
  printf("Add %d: (%d) ", u.v, a->p - current_block);
  printvs(a);
  #endif

  if (a->p - current_block == VERTEX_SET_BLOCK_SIZE)
  {
    allocateNewBlock();
    long diff = (long) (current_block + 2 - a);
    PVertexSet cpy;
    for (cpy = a + 1; cpy < a->p; cpy++)
      *(cpy + diff) = *cpy;
    
    current_block[2].p = current_block + 3 + getSize(a);
    a = current_block + 2;
  }
  
  a->p++;
  PVertexSet first = a;
  a = a->p - 2;
  while (a->v > u.v && a > first)
    *(a + 1) = *a--;
  *(a + 1) = u;
  #ifdef DEBUG
  printf("After: ");
  printvs(first);
  #endif
  
  return first;
}

void removeVertex(PVertex u, PVertexSet a)
{
  #ifdef DEBUG
  printf("Rmv %d: ", u);
  printvs(a);
  #endif
  PVertexSet t = a;
  PVertexSet last = a->p;
  a++;
  while (a->v < u.v)
    a++;
  while (a < last)
    *a++ = *(a + 1);
  t->p--;
  #ifdef DEBUG
  printf("After: ");
  printvs(t);
  #endif
}

PVertexSet setUnion(PVertexSet a, PVertexSet b)
{
  #ifdef DEBUG
  printf("Union: (%d)", a->p + getSize(b) - current_block);
  printvs(a);
  printvs(b);
  #endif
  
  if (a->p + getSize(b) - current_block >= VERTEX_SET_BLOCK_SIZE)
  {
    allocateNewBlock();
    long diff = (long) (current_block + 2 - a);
    PVertexSet cpy;
    for (cpy = a + 1; cpy < a->p; cpy++)
      *(cpy + diff) = *cpy;
    
    current_block[2].p = current_block + 3 + getSize(a);
    a = current_block + 2;
  }
  
  PVertexSet alast = a->p - 1;
  PVertexSet blast = b->p - 1;
  PVertexSet write = alast + (blast - b);
  PVertexSet wlast = write + 1;
  
  while (a < alast && b < blast)
    if (alast->v < blast->v)
      *write-- = *blast--;
    else
    {
      if (alast->v == blast->v)
        blast--;
      *write-- = *alast--;
    }
  
  #ifdef DEBUG
  a->p = wlast;
  printf("In between (%d %d): ", alast - a, blast - b);
  printvs(a);
  #endif
  
  long diff = write - blast;
  for (; b < blast; blast--)
    *(blast + diff) = *blast;
  write = blast + diff;

  diff = write - alast;
  for (write++; write < wlast; write++)
    *(write - diff) = *write;
  
  a->p = write - diff;
  #ifdef DEBUG
  printf("After: ");
  printvs(a);
  #endif
  
  return a;
}

void setIntersection(PVertexSet a, PVertexSet b)
{
  #ifdef DEBUG
  printf("Int: ");
  printvs(a);
  printvs(b);
  #endif
  PVertexSet t = a;
  PVertexSet alast = a->p;
  PVertexSet blast = b->p;
  a++;
  b++;
  PVertexSet w = a;
  while (a < alast && b < blast)
  {
    if (a->v == b->v)
    {
      *w++ = *a++;
      b++;
    }
    else if (b->v < a->v)
      b++;
    else
      a++;
  }
  t->p = w;
  #ifdef DEBUG
  printf("After: ");
  printvs(t);
  #endif
}

void setMinus(PVertexSet a, PVertexSet b)
{
  #ifdef DEBUG
  printf("Minus: ");
  printvs(a);
  printvs(b);
  #endif
  PVertexSet t = a;
  PVertexSet alast = a->p;
  PVertexSet blast = b->p;
  a++;
  b++;
  PVertexSet w = a;
  while (a < alast && b < blast)
    if (a->v == b->v)
    {
      a++;
      b++;
    }
    else if (b->v < a->v)
      b++;
    else
      *w++ = *a++;
  while (a < alast)
    *w++ = *a++;
  t->p = w;
  #ifdef DEBUG
  printf("After: ");
  printvs(t);
  #endif
}

int containsVertex(PVertex u, PVertexSet a)
{
  #ifdef DEBUG
  printf("Is %d contained in ", u);
  printvs(a);
  #endif
  PVertexSet first = a + 1;
  PVertexSet last = a->p - 1;
  if (first->v == u.v || last->v == u.v)
  {
    #ifdef DEBUG
    printf("It is!\n");
    #endif
    return TRUE;
  }
  
  PVertexSet mid;
  
  while (first + 1 < last)
  {
    mid = first + ((last - first) >> 1);
    if (mid->v > u.v)
      last = mid;
    else if (mid->v < u.v)
      first = mid;
    else
    {
      #ifdef DEBUG
      printf("It is!\n");
      #endif
      return TRUE;
    }
  }

  #ifdef DEBUG
  printf("It isn't.\n");
  #endif
  return FALSE;
}

int setEqual(PVertexSet a, PVertexSet b)
{
  #ifdef DEBUG
  printf("Does ");
  printvs(a);
  printf("equal ");
  printvs(b);
  #endif
  PVertexSet alast = a->p;
  PVertexSet blast = b->p;
  if (alast - a != blast - b)
  {
    #ifdef DEBUG
    printf("It doesn't.\n");
    #endif
    return FALSE;
  }
  
  a++;
  b++;
  while (a->v == b->v && a < alast)
  {
    a++;
    b++;
  }
  
  #ifdef DEBUG
  printf("It %s\n", a == alast?"does!":"doesn't.");
  #endif
  return a == alast;
}

int subset(PVertexSet a, PVertexSet b)
{
  #ifdef DEBUG
  printf("Is ");
  printvs(a);
  printf("a subset of ");
  printvs(b);
  #endif
  PVertexSet alast = a->p;
  PVertexSet blast = b->p;
  if (alast - a > blast - b)
  {
    #ifdef DEBUG
    printf("It isn't.\n");
    #endif
    return FALSE;
  }
  
  a++;
  b++;
  while (a < alast && b < blast)
  {
    if (a->v >= b->v)
    {
      if (a->v == b->v)
        a++;
      b++;
    }
    else
    {
      #ifdef DEBUG
      printf("It isn't.\n");
      #endif
      return FALSE;
    }
  }
  #ifdef DEBUG
  printf("It %s\n", a == alast?"is!":"isn't.");
  #endif
  return a == alast;
}

PVertexSetIterator* getVertexIterator(PVertexSet a)
{
  PVertexSetIterator* iter = (PVertexSetIterator*) malloc(sizeof(PVertexSetIterator));
  if (iter == NULL)
  {
    perror("Error allocating vertex set iterator: ");
    exit(-1);
  }
  
  iter->end = a->p;
  iter->next = a + 1;
  return iter;
}

PVertex getNext(PVertexSetIterator* iter)
{
  PVertex ret = *iter->next;
  iter->next++;
  return ret;
}

#endif
